import enum
import itertools

import scipy.linalg

from .VectorOps import vec_normalize, vec_angles
from . import VectorOps as vec_ops
from . import Misc as misc
from . import PermutationOps as perm_ops
import math, numpy as np, scipy as sp

__all__ = [
    "rotation_matrix",
    "skew_symmetric_matrix",
    "rotation_matrix_skew",
    "youla_skew_decomp",
    "youla_skew_matrix",
    "youla_angles",
    "youla_matrix",
    "skew_from_rotation_matrix",
    "translation_matrix",
    "affine_matrix",
    "reflection_matrix",
    "permutation_matrix",
    "extract_rotation_angle_axis",
    "extract_reflection_axis",
    "view_matrix",
    "find_coordinate_matching_permutation",
    "symmetry_permutation",
    "apply_symmetries",
    "symmetry_reduce",
    "identify_cartesian_transformation_type",
    "TransformationTypes",
    "cartesian_transformation_from_data"
]

#######################################################################################################################
#
#                                                 rotation_matrix
#

def rotation_matrix_2d(theta):
    return np.moveaxis(
        np.array([
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]
        ]),
        0, -2
    )

def rotation_matrix_basic(xyz, theta):
    """rotation matrix about x, y, or z axis

    :param xyz: x, y, or z axis
    :type xyz: str
    :param theta: counter clockwise angle in radians
    :type theta: float
    """

    axis = xyz.lower()
    theta = np.asanyarray(theta)
    one = np.ones_like(theta)
    zero = np.zeros_like(theta)
    if axis == "z": # most common case so it comes first
        mat = [
            [ np.cos(theta), -np.sin(theta), zero],
            [ np.sin(theta),  np.cos(theta), zero],
            [ zero,                    zero,  one]
        ]
    elif axis == "y":
        mat = [
            [ np.cos(theta), zero, -np.sin(theta)],
            [ zero,           one,           zero],
            [ np.sin(theta), zero,  np.cos(theta)]
        ]
    elif axis == "x":
        mat = [
            [  one,           zero,           zero],
            [ zero,  np.cos(theta), -np.sin(theta)],
            [ zero,  np.sin(theta),  np.cos(theta)]
        ]
    else:
        raise Exception("{}: axis '{}' invalid".format('rotation_matrix_basic', xyz))
    return np.moveaxis(np.array(mat), 0, -1)

#thank you SE for the nice Euler-Rodrigues imp: https://stackoverflow.com/questions/6802577/rotation-of-3d-vector
def rotation_matrix_ER(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([
        [aa + bb - cc - dd, 2 * (bc + ad),     2 * (bd - ac)    ],
        [2 * (bc - ad),     aa + cc - bb - dd, 2 * (cd + ab)    ],
        [2 * (bd + ac),     2 * (cd - ab),     aa + dd - bb - cc]
    ])

def rotation_matrix_ER_vec(axes, thetas):
    """
    Vectorized version of basic ER
    """

    axes = vec_normalize(np.asanyarray(axes))
    thetas = np.asanyarray(thetas)
    # if len(axes.shape) == 1:
    #     axes = axes/np.linalg.norm(axes)
    #     axes = np.broadcast_to(axes, (len(thetas), 3))
    # else:
    #     axes = vec_normalize(axes)

    ax_shape = axes.shape[:-1]
    t_shape = thetas.shape
    if thetas.ndim == 0:
        base_shape = ax_shape
    elif axes.ndim == 1:
        base_shape = t_shape
    elif thetas.ndim != axes.ndim - 1:
        raise ValueError(f"can't broadcast axes and angles with shapes {ax_shape} and {t_shape}")
    else:
        base_shape = tuple(a if t == 1 else t for a,t in zip(ax_shape, t_shape))

    axes = np.reshape(axes, (-1, 3))
    thetas = thetas.reshape(-1)
    if thetas.shape[0] < axes.shape[0]:
        thetas = np.broadcast_to(thetas, axes.shape[:-1])
    elif axes.shape[0] < thetas.shape[0]:
        axes = np.broadcast_to(axes, thetas.shape + axes.shape[-1:])

    a = np.cos(thetas/2.0)
    b, c, d = np.moveaxis(-axes * np.reshape(np.sin(thetas / 2.0), (len(thetas), 1)), -1, 0)
    v = np.array([a, b, c, d])
    # triu_indices
    rows, cols = (
        np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 3]),
        np.array([0, 1, 2, 3, 1, 2, 3, 2, 3, 3])
    )
    aa, ab, ac, ad, bb, bc, bd, cc, cd, dd = v[rows] * v[cols]
    ## Uses half-angle formula to get compact form for Euler-Rodrigues
    # a^2 * I + [[ 0,    2ad, -2ac]   + [[b^2 - c^2 - d^2,               2bc,              2bd]
    #            [-2ad,    0,  2ab],     [             2bc, -b^2 + c^2 - d^2,              2cd]
    #            [ 2ac, -2ab,    0]]     [             2bd,              2cd, -b^2 - c^2 + d^2]]
    R = np.array([
        [aa + bb - cc - dd,      2 * (bc + ad),         2 * (bd - ac)],
        [    2 * (bc - ad),  aa - bb + cc - dd,         2 * (cd + ab)],
        [    2 * (bd + ac),      2 * (cd - ab),     aa - bb - cc + dd]
    ])
    R = np.moveaxis(R, -1, 0)

    return R.reshape(base_shape + (3, 3))

def rotation_matrix_align_vectors(vec1, vec2):
    vec1 = vec_ops.vec_normalize(vec1)
    vec2 = vec_ops.vec_normalize(vec2)
    s = vec1 + vec2
    i = vec_ops.identity_tensors(vec1.shape[:-1], vec1.shape[-1])
    inner = 1 + vec1[..., np.newaxis, :] @ vec2[..., :, np.newaxis]

    return i - s[..., :, np.newaxis] * (s[..., np.newaxis, :]/inner) + 2 * vec1[..., :, np.newaxis] * vec2[..., np.newaxis, :]

def rotation_matrix(axis, theta=None):
    """
    :param axis:
    :type axis:
    :param theta: angle to rotate by (or Euler angles)
    :type theta:
    :return:
    :rtype:
    """
    if theta is None:
        skew_vector = np.asanyarray(axis, dtype=float)
        if skew_vector.ndim == 0:
            skew_vector = skew_vector[np.newaxis]
        base_shape = skew_vector.shape[:-1]
        skew_vector = skew_vector.reshape(-1, skew_vector.shape[-1])
        rots = np.array([
            rotation_matrix_skew(v)
            for v in skew_vector
        ])
        return rots.reshape(base_shape + rots.shape[-2:])
    elif isinstance(axis, str):
        if axis.lower() == '2d':
            return rotation_matrix_2d(theta)

        theta = np.asanyarray(theta)
        if len(axis) == 1 and theta.ndim == 1 or theta.shape[-2] != 1:
            theta = np.expand_dims(theta, -2)
        rot_mat = None
        for x,t in zip(axis.lower(), np.moveaxis(theta, -2, 0)):
            rr = rotation_matrix_basic(x, t)
            if rot_mat is None:
                rot_mat = rr
            else:
                rot_mat = rr @ rot_mat
        return rot_mat
    else:
        axis = np.asanyarray(axis)
        theta = np.asanyarray(theta)

        if axis.shape == theta.shape:
            return rotation_matrix_align_vectors(axis, theta)
        elif (axis.ndim == theta.ndim + 1) or theta.ndim == 0 and axis.shape[-1] == 3:
            return rotation_matrix_ER_vec(axis, theta)
        else:
            # we have the vectors that get mixed and their mixing angles, we assume any fixed axis is the 0 element
            # I haven't vectorized this at all...
            base_shape = axis.shape[:-1]
            theta = theta.reshape(-1, theta.shape[-1])
            axis = axis.reshape((-1,) + axis.shape[-1:])

            rots = np.array([
                rotation_matrix_from_angles_vectors(l, T)
                for T, l in zip(axis, theta)
            ])

            return rots.reshape(base_shape + rots.shape[-2:])

def skew_symmetric_matrix(upper_tri):
    upper_tri = np.asanyarray(upper_tri)
    l = upper_tri.shape[-1]
    n = (1 + np.sqrt(1 + 8*l)) / 2
    if int(n) != n:
        raise ValueError(f"vector of shape {l} doesn't correspond to the upper triangle of a matrix")
    n = int(n)
    base_shape = upper_tri.shape[:-1]
    m = np.zeros(base_shape + (n, n))
    rows, cols = np.triu_indices_from(m, 1)
    m[..., rows, cols] = upper_tri
    m[..., cols, rows] = -upper_tri
    return m

def extract_rotation_angle_axis(rot_mat, normalize=True):
    rot_mat = np.asanyarray(rot_mat)
    if rot_mat.shape[-1] == 2:
        return np.arccos(rot_mat[..., 0, 0]), None
    elif rot_mat.shape[-1] == 3:
        base_shape = rot_mat.shape[:-2]
        rot_mat = rot_mat.reshape((-1,) + rot_mat.shape[-2:])
        rows, cols = (np.array([2, 0, 1]), np.array([1, 2, 0]))
        skew = (rot_mat[:, rows, cols] - rot_mat[:, cols, rows]) / 2

        ax = np.empty(rot_mat.shape[:-1], dtype=float)
        ang = np.empty(rot_mat.shape[:-2], dtype=float)

        pos_mask = np.linalg.norm(skew, axis=-1) < 1e-6
        bad_pos = np.where(pos_mask)[0]
        rem_pos = np.arange(len(rot_mat))
        if len(bad_pos) > 0:
            rem_pos = np.setdiff1d(rem_pos, bad_pos)
            shift_rot = 1/2 * (rot_mat[bad_pos,] + vec_ops.identity_tensors((len(bad_pos),), rot_mat.shape[-1]))
            for b,r in zip(bad_pos, shift_rot):
                ord = np.lexsort(r)
                ax[b] = r[ord[-1]]
            if normalize:
                ax[bad_pos,] = vec_ops.vec_normalize(ax[bad_pos,])

            # check identity
            id_mask = np.abs(np.trace(rot_mat[bad_pos,], axis1=-1, axis2=-2) > 3-3e-6)
            ang[bad_pos,] = np.pi * np.logical_not(id_mask)

        if len(rem_pos) > 0:
            ax[rem_pos,] = vec_ops.vec_normalize(skew[rem_pos,])

            # uses an efficient implementation by Jens Noeckel: https://mathematica.stackexchange.com/a/29966/38205
            proj = vec_ops.orthogonal_projection_matrix(ax[rem_pos,][:, :, np.newaxis])
            perm = skew[rem_pos,][:, (1, 2, 0)] * np.array([-1, 1, 1])
            ort_vec = (proj @ perm[:, :, np.newaxis]).reshape(perm.shape)
            ort_norms = np.linalg.norm(ort_vec, axis=-1)
            bad_orts = np.where(ort_norms < 1e-4)
            if len(bad_orts) > 0 and len(bad_orts[0]) > 0:
                perm[bad_orts] = vec_ops.vec_normalize(np.random.uniform(size=perm[bad_orts].shape))
                ort_vec = (proj @ perm[:, :, np.newaxis]).reshape(perm.shape)
            normal = vec_ops.vec_crosses(ax[rem_pos,], ort_vec)
            w = np.reshape(rot_mat[rem_pos,] @ ort_vec[:, :, np.newaxis], ort_vec.shape)
            ang[rem_pos,] = np.arctan2(vec_ops.vec_dots(w, normal), vec_ops.vec_dots(w, ort_vec))


        # """
        # ovec = w - axis Dot[w, axis];
        # nvec = Cross[axis, ovec];
        # w1 = m . ovec;
        # """
        #
        # tr_rot = np.trace(rot_mat, axis1=-2, axis2=-1)
        # ang = np.arccos((tr_rot - 1) / 2)
        #
        # zero_mask = np.abs(ang) < 1e-6
        # zero_pos = np.where(zero_mask)
        # gimbal_mask = np.abs(np.abs(ang) - np.pi) < 1e-6
        # gimbal_locked = np.where(gimbal_mask)
        # rem = np.where(np.logical_not(np.logical_or(zero_mask, gimbal_mask)))
        #
        # ax = np.zeros(rot_mat.shape[:-1])
        # if len(zero_pos) > 0 and len(zero_pos[0]) > 0:
        #     ax[zero_pos] = np.repeat([[0, 0, 1]], len(zero_pos[0]), axis=0)
        # if len(gimbal_locked) > 0:
        #     eigs, gimb_ax = np.linalg.eigh(rot_mat[gimbal_locked])
        #     one_pos = np.where(eigs > 1-1e-4)
        #     ax[gimbal_locked] = gimb_ax[one_pos[0], :, one_pos[1]]
        # if len(rem) > 0 and len(rem[0]) > 0:
        #     skew = (rot_mat[rem] - np.moveaxis(rot_mat[rem], -1, -2)) / 2
        #     rows, cols = (np.array([2, 0, 1]), np.array([1, 2, 0]))
        #     ax[rem] = skew[..., rows, cols]
        #     ax[rem] = skew[..., rows, cols]
        # if normalize:
        #     ax = vec_normalize(ax)
        # else:
        #     ax = ax
        return ang.reshape(base_shape), ax.reshape(base_shape + ax.shape[-1:])
    else:
        base_shape = rot_mat.shape[:-2]
        rot_mat = np.reshape(rot_mat, (-1,) + rot_mat.shape[-2:])
        angles = []
        axes = []
        for r in rot_mat:
            U, Q = scipy.linalg.schur(r)
            angles.append(youla_angles(U))
            axes.append(Q)

        angles = np.array(angles)
        axes = np.array(axes)

        return angles.reshape(base_shape + angles.shape[-1:]), axes.reshape(base_shape + axes.shape[-2:])

def extract_reflection_axis(reflection_mat):
    reflection_mat = np.asarray(reflection_mat)
    base_shape = reflection_mat.shape[:-2]
    reflection_mat = np.reshape(reflection_mat, (-1,) + reflection_mat.shape[-2:])
    # if reflection_mat.shape[-1] < 3:
    pt = vec_ops.vec_normalize(np.random.rand(1, reflection_mat.shape[-1]), axis=-1)
    tf_pts = np.reshape(reflection_mat @ pt[:, :, np.newaxis], (reflection_mat.shape[0], reflection_mat.shape[-1]))
    vecs = (pt - tf_pts)
    norms = vec_ops.vec_norms(vecs)
    in_planes = np.where(norms < 1e-4)
    if len(in_planes) > 0 and len(in_planes[0]) > 0: # might have randomly ended up in the reflection plane
        subpt = pt[in_planes]
        subref = reflection_mat[in_planes]
        pt2 = vec_ops.vec_normalize(np.random.rand(1, reflection_mat.shape[-1]), axis=-1)
        pt2 = pt2 - vec_ops.vec_dots(pt[in_planes], pt2)[:, np.newaxis]*pt[in_planes]
        tf_pts = np.reshape(subref @ subpt[:, :, np.newaxis], (subref.shape[0], reflection_mat.shape[-1]))
        vecs[in_planes] = (pt2 - tf_pts)
        norms[in_planes] = vec_ops.vec_norms(vecs[in_planes])
    return vec_ops.vec_normalize(vecs, norms).reshape(base_shape + (reflection_mat.shape[-1],))

def youla_skew_decomp(A):
    n = len(A)
    s, T = sp.linalg.schur(A)

    l = np.diag(s, 1)
    if n % 2 == 0:
        start = 0
        end = n
    else:  # manage padding for odd dimension
        if abs(l[0]) < 1e-7:
            start = 1
            end = n
        else:
            start = 0
            end = n - 1
    l = l[start:end-1:2]

    return youla_matrix(l, n, axis_pos=0 if start == 0 else n), T

def youla_skew_matrix(l, n, axis_pos=0):

    U = np.zeros((n, n))
    o = np.concatenate([  # even inds
        np.arange(0, axis_pos, 2),
        np.arange(axis_pos + 1, n, 2),
    ])
    e = np.concatenate([  # odd inds
        np.arange(1, axis_pos, 2),
        np.arange(axis_pos + 2, n, 2),
    ])

    U[o, e] = l
    U[e, o] = -l

    return U

def youla_matrix(angles, n, axis_pos=0):

    cos = np.cos(angles)
    sin = np.sin(angles)

    # build 2x2 block mat
    U = np.eye(n)
    if n % 2 == 1:
        o = np.concatenate([  # even inds
            np.arange(0, axis_pos, 2),
            np.arange(axis_pos+1, n, 2),
            ])
        e = np.concatenate([  # even inds
            np.arange(1, axis_pos, 2),
            np.arange(axis_pos+2, n, 2),
            ])
    else:
        o = np.arange(0, n, 2)
        e = np.arange(1, n, 2)

    U[o, o] = cos
    U[e, e] = cos
    U[o, e] = sin
    U[e, o] = -sin

    return U

def youla_angles(U, axis_pos=None):
    l = np.arccos(np.round(np.diag(U), 8))
    n = len(U)
    if axis_pos is None:
        if n % 2 == 0:
            axis_pos = -1
        else:  # manage padding for odd dimension
            axis_pos = np.where(abs(l) < 1e-7)[0]
            axis_pos = 0 if axis_pos[0] == 0 else axis_pos[-1]
        if axis_pos < 0:
            axis_pos = n + axis_pos

    return np.concatenate([
        l[0:axis_pos:2],
        l[axis_pos+1::2]
    ])

def rotation_matrix_skew(upper_tri, create_skew=True):
    upper_tri = np.asanyarray(upper_tri)
    if create_skew:
        if (
                upper_tri.ndim < 2
                or upper_tri.shape[-1] != upper_tri.shape[-2]
                or not np.allclose(upper_tri, -np.moveaxis(upper_tri, -2, -1))
        ):
            upper_tri = skew_symmetric_matrix(upper_tri)

    U, T = youla_skew_decomp(upper_tri)
    return T@U@T.T

def skew_from_rotation_matrix(rot_mat):
    U, Q = sp.linalg.schur(rot_mat)

    l = youla_angles(U)
    s = youla_skew_matrix(l, U.shape[0], axis_pos=0 if U[0, 0] == 1 else U.shape[0])
    A = Q @ s @ Q.T
    return A[np.triu_indices_from(A, k=1)]

def rotation_matrix_from_angles_vectors(l, T):
    n = T.shape[0]
    if n % 2 == 1 and len(l) == (n // 2) + 1: # the axis is encoded in l
        axis_pos = np.where(np.abs(l) > 2 * np.pi)[0]
        if len(axis_pos) == 0:
            axis_pos = np.where(np.abs(l) < 1e-7)[0]
            if len(axis_pos) == 0:
                raise ValueError(f"can't find fixed axis position from angle encoding {l}")
        axis_pos = axis_pos[0]
    else:
        axis_pos = 0
    return T @ youla_matrix(l, n, axis_pos=axis_pos) @ T.T

#######################################################################################################################
#
#                                                 translation_matrix
#

def translation_matrix(shift):
    share = np.asarray(shift)
    if len(share.shape) == 1:
        ss = share
        zs = 0.
        os = 1.
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        )
    else:
        zs = np.zeros((share.shape[0],))
        os = np.ones((share.shape[0],))
        ss = share.T
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        ).T
    return mat

#######################################################################################################################
#
#                                                 affine_matrix
#

def affine_matrix(tmat, shift):
    """Creates an affine transformation matrix from a 3x3 transformation matrix or set of matrices and a shift or set of vecs

    :param tmat: base transformation matrices
    :type tmat: np.ndarray
    :param shift:
    :type shift:
    :return:
    :rtype:
    """

    base_mat = np.asanyarray(tmat)
    if shift is None:
        return base_mat

    if base_mat.ndim > 2:
        shifts = np.asanyarray(shift)
        if shifts.ndim == 1:
            shifts = np.broadcast_to(shifts, (1,)*(base_mat.ndim-2) + shifts.shape)
        shifts = np.broadcast_to(shifts, base_mat.shape[:-2] + (shifts.shape[-1],))
        shifts = np.expand_dims(shifts, -1)
        mat = np.concatenate([base_mat, shifts], axis=-1)
        padding = np.zeros(shifts.shape[-2]+1)
        padding[-1] = 1
        padding = np.broadcast_to(
            np.broadcast_to(padding, (1,)*(base_mat.ndim-2) + padding.shape),
            mat.shape[:-2] + (shifts.shape[-2]+1,)
        )
        padding = np.expand_dims(padding, -2)
        mat = np.concatenate([mat, padding], axis=-2)
    else:
        shift = np.asanyarray(shift)
        mat = np.concatenate([base_mat, shift[:, np.newaxis]], axis=-1)
        padding = np.zeros(shift.shape[-1]+1)
        padding[-1] = 1
        mat = np.concatenate([mat, padding[np.newaxis]], axis=-2)
    return mat

def view_matrix(
        up_vector,
        view_vector=(0, 0, 1),
        output_order=None
):

    up_vector = vec_ops.vec_normalize(up_vector)
    d = up_vector.shape[-1]
    base_shape = up_vector.shape[:-1]
    up_vector = up_vector.reshape(-1, d)
    view_vector = vec_normalize(view_vector)
    view_vector = view_vector.reshape(-1, d)
    overlaps = vec_ops.vec_dots(up_vector, view_vector)
    bad_views = np.where(np.abs(overlaps) > 1-1e-6) # TOOD: make this a threshold
    if len(bad_views) > 0 and len(bad_views[0]) > 0:
        new_views = np.repeat(np.array([[1, 0, 0]]), len(bad_views[0])).copy()
        overlaps2 = vec_ops.vec_dots(new_views, view_vector[bad_views])
        bad_views2 = np.where(np.abs(overlaps2) > 1-1e-6) # TOOD: make this a threshold
        if len(bad_views2) > 0 and len(bad_views2[0]) > 0:
            new_views[bad_views2] = np.array([[0, 0, 1]])
        new_views = new_views - vec_ops.vec_dots(new_views, view_vector[bad_views]) * view_vector[bad_views]
        view_vector[bad_views] = new_views
    right_vector = vec_ops.vec_normalize(
        vec_ops.vec_crosses(up_vector, view_vector)
    )
    view_vector = vec_ops.vec_normalize(
        vec_ops.vec_crosses(up_vector, right_vector)
    )
    axes = np.concatenate([
        view_vector[..., np.newaxis],
        up_vector[..., np.newaxis],
        right_vector[..., np.newaxis]
    ], axis=-1).reshape(base_shape + (d, d))

    if output_order is not None:
        from .PermutationOps import permutation_sign
        sign = permutation_sign(output_order)
        axes = axes[..., :, output_order]
        axes[..., :, 1] *= sign

    return axes

def reflection_matrix(axes):
    # need to find space of "null" vectors
    axes = np.asanyarray(axes)
    smol = axes.ndim == 1
    if smol:
        axes = axes[np.newaxis]
    if axes.ndim == 2:
        axes = axes[:, np.newaxis, :]
    base_shape = axes.shape[:-2]
    axes = axes.reshape((-1,) + axes.shape[-2:])
    nax = axes.shape[-2]
    ndim = axes.shape[-1]
    eyes = vec_ops.identity_tensors(axes.shape[0], ndim)
    full_basis = np.concatenate([axes, eyes], axis=-2)
    q, r = np.linalg.qr(np.moveaxis(full_basis, -2, -1))
    tf = q[:, :, :ndim]

    diag_refl = np.diag([-1] * nax + [1] * (ndim - nax))[np.newaxis]
    refls = tf @ diag_refl @ np.moveaxis(tf, -2, -1)

    refls = refls.reshape(base_shape + refls.shape[1:])
    if smol: refls = refls[0]

    return refls

def permutation_matrix(perm):
    perm = np.asanyarray(perm)
    smol = perm.ndim == 1
    if smol:
        perm = perm[np.newaxis]
    base_shape = perm.shape[:-1]
    perm = perm.reshape(-1, perm.shape[-1])
    mats = np.zeros((perm.shape[0], perm.shape[-1], perm.shape[-1]), dtype='uint8')
    ix = (
        np.expand_dims(np.arange(mats.shape[0]), [1]),
        np.expand_dims(np.arange(perm.shape[-1]), [0]),
        perm
    )
    mats[ix] = 1

    mats = mats.reshape(base_shape + mats.shape[1:])
    if smol: mats = mats[0]

    return mats

def find_coordinate_matching_permutation(coords, new_coords, return_row_ordering=False, tol=None):
    base_shape = coords.shape[:-2]
    coords = coords.reshape((-1,) + coords.shape[-2:])
    new_coords = new_coords.reshape((-1,) + new_coords.shape[-2:])
    dm = np.linalg.norm(coords[:, :, np.newaxis] - new_coords[:, np.newaxis, :], axis=-1)
    # quick_test = np.where(dm < 1e-2)
    # iteratively find best matches and map them onto each other
    perm_rows = np.repeat(np.arange(coords.shape[1])[np.newaxis], coords.shape[0], axis=0)
    perm_cols = np.repeat(np.arange(coords.shape[1])[np.newaxis], coords.shape[0], axis=0)
    a = np.arange(dm.shape[0])
    a1 = np.arange(dm.shape[0])[:, np.newaxis]
    a2 = np.arange(dm.shape[1])
    for i in range(coords.shape[1]):
        # we drop one more row each loop
        if i > 0:
            sel = a2[np.newaxis, :(-i)]
        else:
            sel = a2[np.newaxis, :]

        # print(np.round(dm, 8))
        min_pos = np.argmin(dm, axis=-1) # min across columns
        min_vals = dm[a1, sel, min_pos]
        ord = np.argsort(min_vals, axis=-1) # min across rows
        if tol is not None:
            max_min = np.max(min_vals)
            if max_min > tol:
                raise ValueError(f"maximum minimum deviation too large {max_min} > {tol}")
        # ord = np.argsort(ord, axis=-1)
        # assign row permutation
        perm_next = perm_rows[:, i:][a1[:, np.newaxis], ord[:, 0]]
        perm_rows[:, i+1:] = perm_rows[:, i:][a1[:, np.newaxis], ord[:, 1:]]
        perm_rows[:, i] = perm_next

        # # column can be duplicated, so we reorder just based on the minimum row
        min_pos = np.argsort(dm[a, ord[:, 0]], axis=-1) # min across columns
        # # print("?", ord[:, 0].shape, dm[a1, ord[:, 0]].shape)
        # # min_pos = min_pos[a1, ord]
        perm_next = perm_cols[:, i:][a1[:, np.newaxis], min_pos[:, 0]]
        # print(perm_cols[:, i:][a1[:, np.newaxis], min_pos[:, 1:]])
        perm_cols[:, i+1:] = perm_cols[:, i:][a1[:, np.newaxis], min_pos[:, 1:]]
        perm_cols[:, i] = perm_next

        # print("?", perm_rows)
        # print(">", np.sort(min_pos[:, 1:], axis=-1))
        # print(ord)
        dm = dm[a1, ord[:, 1:, np.newaxis], min_pos[:, np.newaxis, 1:]]


    if return_row_ordering:
        return perm_rows.reshape(base_shape), perm_cols.reshape(base_shape)
    else:
        row_sort = np.argsort(perm_rows, axis=1)
        perms = perm_cols[a1, row_sort].reshape(base_shape + perm_cols.shape[1:])
        return perms

def symmetry_permutation(coords, op:np.ndarray, return_row_ordering=False, tol=None):
    # converts a symmetry operation into a permutation of the coords
    coords = np.asanyarray(coords)
    op = np.asanyarray(op)

    base_shape = coords.shape[:-2]
    coords = coords.reshape((-1,) + coords.shape[-2:])
    op = op.reshape((-1,) + op.shape[-2:])
    new_coords = coords @ op

    perm_data = find_coordinate_matching_permutation(coords, new_coords,
                                                     return_row_ordering=return_row_ordering, tol=tol
                                                     )

    if return_row_ordering:
        perm_rows, perm_cols = perm_data
        return perm_rows.reshape(base_shape + perm_rows.shape[-1:]), perm_cols.reshape(base_shape + perm_cols.shape[-1:])
    else:
        return perm_data.reshape(base_shape + perm_data.shape[-1:])

def apply_symmetries(coords, symmetry_elements: 'list[np.ndarray]', labels=None, tol=1e-1):
    coords = np.asanyarray(coords)
    if coords.ndim == 1:
        coords = coords[np.newaxis]
    elif coords.ndim > 2:
        raise NotImplementedError("currently don't have good batching on coord symmetry application")
    og_labels = labels
    if labels is not None:
        if labels is True:
            labels = np.arange(len(coords))
        else:
            labels = np.arange(len(labels))
    # symmetry_elements = prep_symmetry_operations(symmetry_elements)

    for e in symmetry_elements:
        new_coords = coords @ e
        coord_diffs = np.linalg.norm(coords[:, np.newaxis, :] - new_coords[np.newaxis, :, :], axis=-1)
        dupe_pos = np.where(coord_diffs < tol)
        new_labs = None
        if len(dupe_pos) > 0 and len(dupe_pos[0]) > 0:
            rem = np.setdiff1d(np.arange(len(new_coords)), dupe_pos[0])
            if labels is not None:
                new_labs = labels[rem,]
            new_coords = new_coords[rem,]
        elif labels is not None:
            new_labs = labels
        if labels is not None:
            labels = np.concatenate([labels, new_labs])
        coords = np.concatenate([coords, new_coords], axis=0)

    if labels is not None:
        if og_labels is not True:
            labels = [og_labels[i] for i in labels]
        return coords, labels
    else:
        return coords

def symmetry_reduce(coords, op:np.ndarray, labels=None):
    coords = np.asarray(coords)
    if coords.ndim == 1:
        if labels is not None:
            return coords, labels
        else:
            return coords
    elif coords.ndim > 2:
        raise NotImplementedError("currently don't have good batching on coord symmetry reduction")

    perm = symmetry_permutation(coords, op)
    cycles = perm_ops.permutation_cycles(perm, return_groups=True)
    coords = np.array([
        coords[p[0]]
        for p in cycles
    ])
    if labels is not None:
        labels = [
            labels[p[0]]
            for p in cycles
        ]
        return coords, labels
    else:
        return coords

class TransformationTypes(enum.Enum):
    Identity = 0
    Inversion = 1
    Rotation = 2
    Reflection = 3
    ImproperRotation = 4
    Scaling = 5

def identify_cartesian_transformation_type(x, max_rotation_order=None):
        # check unitary
        x = np.asanyarray(x)
        base_shape = x.shape[:-2]
        x = x.reshape((-1,) + x.shape[-2:])
        scalings, x = vec_ops.polar_decomposition(x)

        types = np.zeros(x.shape[0], dtype=int)
        axes = np.zeros((x.shape[0], 3))
        orders = np.full(x.shape[0], -1, dtype=int)
        roots = np.full(x.shape[0], -1, dtype=float)

        rem_pos = np.arange(x.shape[0])
        t = np.trace(x, axis1=-2, axis2=-1)
        inversion_mask = t < -3 + 1e-2
        types[inversion_mask] = 1

        rem = np.logical_not(inversion_mask)
        rem_pos = rem_pos[rem]
        x = x[rem]
        t = t[rem]


        if len(rem_pos) > 0:
            # identity by default
            rem = t < 3 - 1e-2
            rem_pos = rem_pos[rem]
            x = x[rem]


        if len(rem_pos) > 0:
            d = np.linalg.det(x)
            rotation_mask = d > 1-1e-2

            rotations = x[rotation_mask]

            rot_sel = rem_pos[rotation_mask]
            types[rot_sel,] = 2
            ang, ax = extract_rotation_angle_axis(rotations)
            axes[rot_sel,] = ax

            rationals = ang / (2 * np.pi)
            if max_rotation_order is None:
                roots[rot_sel,] = rationals
                orders[rot_sel,] = 1
            else:
                for rational, i in zip(rationals, rot_sel):
                    for o in range(1, max_rotation_order + 1):
                        root = rational * o
                        if abs(root - np.round(root)) < 1e-6:
                            roots[i] = int(np.round(root))
                            orders[i] = o
                            break
                    else:
                        #TODO: don't just terminate...
                        raise ValueError(
                            f"angle ratio {ang} doesn't correspond to a rational number up to order {max_rotation_order} rotations"
                        )

            rem = np.logical_not(rotation_mask)
            x = x[rem]
            rem_pos = rem_pos[rem]

        if len(rem_pos) > 0:
            reflection_mask = np.max(np.max(x - np.moveaxis(x, -1, -2), axis=-1), axis=-1) < 1e-6 # symmetric
            ref_sel = rem_pos[reflection_mask]
            types[ref_sel,] = 3
            axes[ref_sel,] = extract_reflection_axis(x[reflection_mask])

            rem = np.logical_not(reflection_mask)
            x = x[rem]
            rem_pos = rem_pos[rem]

        if len(rem_pos) > 0:
            # it turns out the same axis trick works for improper rotations
            types[rem_pos,] = 4

            ang, ax = extract_rotation_angle_axis(x)
            axes[rem_pos,] = ax

            rationals = ang / (2 * np.pi)
            if max_rotation_order is None:
                roots[rem_pos,] = rationals
                orders[rem_pos,] = 1
            else:
                for rational, i in zip(rationals, rem_pos):
                    for o in range(1, max_rotation_order + 1):
                        root = rational * o
                        if abs(root - np.round(root)) < 1e-6:
                            roots[i] = int(np.round(root))
                            orders[i] = o
                            break
                    else:
                        # TODO: don't just terminate...
                        raise ValueError(
                            f"angle ratio {ang} doesn't correspond to a rational number up to order {max_rotation_order} rotations"
                        )

        if scalings is not None:
            scalings.reshape(base_shape + scalings.shape[-2:])
        types = types.reshape(base_shape)
        axes = axes.reshape(base_shape + axes.shape[-1:])
        roots = roots.reshape(base_shape)
        orders = orders.reshape(base_shape)
        return scalings, types, axes, roots, orders

def cartesian_transformation_from_data(scalings, types, axes, roots, orders):
    if scalings is not None:
        scalings = np.asanyarray(scalings)
        if scalings.ndim == 1:
            scalings = np.diag(scalings)
        elif scalings.shape[-1] != scalings.shape[-2]:
            scalings = vec_ops.vec_tensordiag(scalings)
    types = np.asanyarray(types)
    axes = np.asanyarray(axes)
    roots = np.asanyarray(roots)
    orders = np.asanyarray(orders)

    base_shape = types.shape
    if scalings is not None:
        scalings = np.reshape(scalings, (-1,) + scalings.shape[-2:])
    types = np.reshape(types, (-1,))
    axes = np.reshape(axes, (-1,) + axes.shape[-1:])
    roots = np.reshape(roots, (-1,))
    orders = np.reshape(orders, (-1,))

    unitary_tfs = np.empty((types.shape[0], axes.shape[-1], axes.shape[-1]), dtype=float)
    rem_pos = np.arange(types.shape[0])

    identity_mask = types == 0
    sel = rem_pos[identity_mask]
    if len(sel) > 0:
        unitary_tfs[sel,] = np.eye(3)[np.newaxis]
        rem = np.logical_not(identity_mask)
        rem_pos = rem_pos[rem]

        types = types[rem]
        axes = axes[rem]
        roots = roots[rem]
        orders = orders[rem]

    if len(types) > 0:
        inverse_mask = types == 1
        sel = rem_pos[inverse_mask]
        if len(sel) > 0:
            unitary_tfs[sel,] = -np.eye(3)[np.newaxis]
            rem = np.logical_not(inverse_mask)
            rem_pos = rem_pos[rem]

            types = types[rem]
            axes = axes[rem]
            roots = roots[rem]
            orders = orders[rem]

    if len(types) > 0:
        rotation_mask = types == 2
        sel = rem_pos[rotation_mask]
        if len(sel) > 0:
            angles = 2*np.pi*roots[rotation_mask]/orders[rotation_mask]
            unitary_tfs[sel,] = rotation_matrix_ER_vec(axes[rotation_mask], angles)
            rem = np.logical_not(rotation_mask)
            rem_pos = rem_pos[rem]

            types = types[rem]
            axes = axes[rem]
            roots = roots[rem]
            orders = orders[rem]

    if len(types) > 0:
        reflection_mask = types == 3
        sel = rem_pos[reflection_mask]
        if len(sel) > 0:
            unitary_tfs[sel,] = reflection_matrix(axes[reflection_mask,])
            rem = np.logical_not(reflection_mask)
            rem_pos = rem_pos[rem]

            types = types[rem]
            axes = axes[rem]
            roots = roots[rem]
            orders = orders[rem]

    if len(types) > 0:
        # improt_mask = types == 4
        angles = 2*np.pi*roots/orders
        unitary_tfs[rem_pos,] = reflection_matrix(axes) @ rotation_matrix_ER_vec(axes, angles)
        # rem = np.logical_not(reflection_mask)
        # rem_pos = rem_pos[rem]
        #
        # types = types[rem]
        # axes = axes[rem]
        # orders = orders[rem]

    if scalings is not None:
        unitary_tfs = scalings @ unitary_tfs

    return unitary_tfs.reshape(base_shape + unitary_tfs.shape[-2:])






"""
Defines a set of useful interactive tools
for working in Jupyter (primarily JupterLab) environments
"""

__all__ = []

from .InteractiveTools import *; from .InteractiveTools import __all__ as exposed
__all__ += exposed
from .JHTML import JHTML
__all__ += ["JHTML"]
from .Apps import *; from .Apps import __all__ as exposed
__all__ += exposed
from .APIs import *; from .APIs import __all__ as exposed
__all__ += exposed
from .ScriptRunner import *; from .ScriptRunner import __all__ as expose
__all__ += exposed
from .MoleculeGraphics import *; from .MoleculeGraphics import __all__ as exposed
__all__ += exposed
from .X3D import *; from .X3D import __all__ as exposed
__all__ += exposed
from .JSMol import *; from .JSMol import __all__ as exposed
__all__ += exposed
from .NotebookTools import *; from .NotebookTools import __all__ as exposed
__all__ += exposed
del exposed

def _ipython_pinfo_():
    from ..Docs import jdoc
    import sys

    return jdoc(sys.modules[__name__])


from .JHTML import HTML

__all__ = [
    "X3DHTML"
]

__reload_hooks__ = [".JHTML"]
class X3DHTML:

    class X3DElement(HTML.TagElement):
        ignored_styles = {"height", "width", "position", "color"}
        style_props = None

        @classmethod
        def convert_attrs(cls, attrs:dict):
            copied = False
            for k,v in attrs.items():
                if isinstance(v, str):
                    continue
                if not copied:
                    copied = True
                    attrs = attrs.copy()
                if v is None:
                    del attrs[k]
                else:
                    if hasattr(v, "__getitem__") or hasattr(v, "__iter__"):
                        v = " ".join(str(x) for x in v)
                    else:
                        v = str(v)
                    attrs[k] = v
            return attrs

        attr_converter = convert_attrs
        # class_map = {"cls":"class_", "id":"id_", "style":"style_"}
        # def to_x3d(self):
        #     tag = x3d.method(self.tag)
        #     opts = {
        #         self.class_map.get(k, k): v.to_x3d() if isinstance(v, X3D.X3DElement) else v
        #         for k, v in self.attrs.items()
        #     }
        #     children = []
        #     for e in self.elems:
        #         e = e.to_x3d() if isinstance(e, X3D.X3DElement) else e
        #         if e.tag in {"head", "Scene"}:
        #             opts[e.tag] = e
        #         else:
        #             children.append(e)
        #
        #     return tag(children=children, **opts)

    class X3D(X3DElement): tag = "X3D"
    class Anchor(X3DElement): tag = "Anchor"
    class Appearance(X3DElement): tag = "Appearance"
    class Arc2D(X3DElement): tag = "Arc2D"
    class ArcClose2D(X3DElement): tag = "ArcClose2D"
    class AudioClip(X3DElement): tag = "AudioClip"
    class Background(X3DElement): tag = "Background"
    class BallJoint(X3DElement): tag = "BallJoint"
    class Billboard(X3DElement): tag = "Billboard"
    class BinaryGeometry(X3DElement): tag = "BinaryGeometry"
    class BlendMode(X3DElement): tag = "BlendMode"
    class BlendedVolumeStyle(X3DElement): tag = "BlendedVolumeStyle"
    class Block(X3DElement): tag = "Block"
    class BoundaryEnhancementVolumeStyle(X3DElement): tag = "BoundaryEnhancementVolumeStyle"
    class Box(X3DElement): tag = "Box"
    class BufferAccessor(X3DElement): tag = "BufferAccessor"
    class BufferGeometry(X3DElement): tag = "BufferGeometry"
    class BufferView(X3DElement): tag = "BufferView"
    class CADAssembly(X3DElement): tag = "CADAssembly"
    class CADFace(X3DElement): tag = "CADFace"
    class CADLayer(X3DElement): tag = "CADLayer"
    class CADPart(X3DElement): tag = "CADPart"
    class CartoonVolumeStyle(X3DElement): tag = "CartoonVolumeStyle"
    class Circle2D(X3DElement): tag = "Circle2D"
    class ClipPlane(X3DElement): tag = "ClipPlane"
    class CollidableShape(X3DElement): tag = "CollidableShape"
    class Collision(X3DElement): tag = "Collision"
    class CollisionCollection(X3DElement): tag = "CollisionCollection"
    class CollisionSensor(X3DElement): tag = "CollisionSensor"
    class Color(X3DElement): tag = "Color"
    class ColorChaser(X3DElement): tag = "ColorChaser"
    class ColorDamper(X3DElement): tag = "ColorDamper"
    class ColorInterpolator(X3DElement): tag = "ColorInterpolator"
    class ColorMaskMode(X3DElement): tag = "ColorMaskMode"
    class ColorRGBA(X3DElement): tag = "ColorRGBA"
    class CommonSurfaceShader(X3DElement): tag = "CommonSurfaceShader"
    class ComposedCubeMapTexture(X3DElement): tag = "ComposedCubeMapTexture"
    class ComposedShader(X3DElement): tag = "ComposedShader"
    class ComposedTexture3D(X3DElement): tag = "ComposedTexture3D"
    class ComposedVolumeStyle(X3DElement): tag = "ComposedVolumeStyle"
    class Cone(X3DElement): tag = "Cone"
    class Coordinate(X3DElement): tag = "Coordinate"
    class CoordinateDamper(X3DElement): tag = "CoordinateDamper"
    class CoordinateDouble(X3DElement): tag = "CoordinateDouble"
    class CoordinateInterpolator(X3DElement): tag = "CoordinateInterpolator"
    class Cylinder(X3DElement): tag = "Cylinder"
    class CylinderSensor(X3DElement): tag = "CylinderSensor"
    class DepthMode(X3DElement): tag = "DepthMode"
    class DirectionalLight(X3DElement): tag = "DirectionalLight"
    class Dish(X3DElement): tag = "Dish"
    class Disk2D(X3DElement): tag = "Disk2D"
    class DoubleAxisHingeJoint(X3DElement): tag = "DoubleAxisHingeJoint"
    class DynamicLOD(X3DElement): tag = "DynamicLOD"
    class EdgeEnhancementVolumeStyle(X3DElement): tag = "EdgeEnhancementVolumeStyle"
    class ElevationGrid(X3DElement): tag = "ElevationGrid"
    class Environment(X3DElement): tag = "Environment"
    class Extrusion(X3DElement): tag = "Extrusion"
    class Field(X3DElement): tag = "Field"
    class FloatVertexAttribute(X3DElement): tag = "FloatVertexAttribute"
    class Fog(X3DElement): tag = "Fog"
    class FontStyle(X3DElement): tag = "FontStyle"
    class GeneratedCubeMapTexture(X3DElement): tag = "GeneratedCubeMapTexture"
    class GeoCoordinate(X3DElement): tag = "GeoCoordinate"
    class GeoElevationGrid(X3DElement): tag = "GeoElevationGrid"
    class GeoLOD(X3DElement): tag = "GeoLOD"
    class GeoLocation(X3DElement): tag = "GeoLocation"
    class GeoMetadata(X3DElement): tag = "GeoMetadata"
    class GeoOrigin(X3DElement): tag = "GeoOrigin"
    class GeoPositionInterpolator(X3DElement): tag = "GeoPositionInterpolator"
    class GeoTransform(X3DElement): tag = "GeoTransform"
    class GeoViewpoint(X3DElement): tag = "GeoViewpoint"
    class Group(X3DElement): tag = "Group"
    class HAnimDisplacer(X3DElement): tag = "HAnimDisplacer"
    class HAnimHumanoid(X3DElement): tag = "HAnimHumanoid"
    class HAnimJoint(X3DElement): tag = "HAnimJoint"
    class HAnimSegment(X3DElement): tag = "HAnimSegment"
    class HAnimSite(X3DElement): tag = "HAnimSite"
    class ImageTexture(X3DElement): tag = "ImageTexture"
    class ImageTexture3D(X3DElement): tag = "ImageTexture3D"
    class ImageTextureAtlas(X3DElement): tag = "ImageTextureAtlas"
    class IndexedFaceSet(X3DElement): tag = "IndexedFaceSet"
    class IndexedLineSet(X3DElement): tag = "IndexedLineSet"
    class IndexedQuadSet(X3DElement): tag = "IndexedQuadSet"
    class IndexedTriangleSet(X3DElement): tag = "IndexedTriangleSet"
    class IndexedTriangleStripSet(X3DElement): tag = "IndexedTriangleStripSet"
    class Inline(X3DElement): tag = "Inline"
    class IsoSurfaceVolumeData(X3DElement): tag = "IsoSurfaceVolumeData"
    class LOD(X3DElement): tag = "LOD"
    class LineProperties(X3DElement): tag = "LineProperties"
    class LineSet(X3DElement): tag = "LineSet"
    class MPRPlane(X3DElement): tag = "MPRPlane"
    class MPRVolumeStyle(X3DElement): tag = "MPRVolumeStyle"
    class Material(X3DElement): tag = "Material"
    class MatrixTextureTransform(X3DElement): tag = "MatrixTextureTransform"
    class MatrixTransform(X3DElement): tag = "MatrixTransform"
    class Mesh(X3DElement): tag = "Mesh"
    class MetadataBoolean(X3DElement): tag = "MetadataBoolean"
    class MetadataDouble(X3DElement): tag = "MetadataDouble"
    class MetadataFloat(X3DElement): tag = "MetadataFloat"
    class MetadataInteger(X3DElement): tag = "MetadataInteger"
    class MetadataSet(X3DElement): tag = "MetadataSet"
    class MetadataString(X3DElement): tag = "MetadataString"
    class MotorJoint(X3DElement): tag = "MotorJoint"
    class MovieTexture(X3DElement): tag = "MovieTexture"
    class MultiTexture(X3DElement): tag = "MultiTexture"
    class MultiTextureCoordinate(X3DElement): tag = "MultiTextureCoordinate"
    class NavigationInfo(X3DElement): tag = "NavigationInfo"
    class NodeNameSpace(X3DElement): tag = "NodeNameSpace"
    class Normal(X3DElement): tag = "Normal"
    class NormalInterpolator(X3DElement): tag = "NormalInterpolator"
    class Nozzle(X3DElement): tag = "Nozzle"
    class OpacityMapVolumeStyle(X3DElement): tag = "OpacityMapVolumeStyle"
    class OrientationChaser(X3DElement): tag = "OrientationChaser"
    class OrientationDamper(X3DElement): tag = "OrientationDamper"
    class OrientationInterpolator(X3DElement): tag = "OrientationInterpolator"
    class OrthoViewpoint(X3DElement): tag = "OrthoViewpoint"
    class Param(X3DElement): tag = "Param"
    class ParticleSet(X3DElement): tag = "ParticleSet"
    class PhysicalEnvironmentLight(X3DElement): tag = "PhysicalEnvironmentLight"
    class PhysicalMaterial(X3DElement): tag = "PhysicalMaterial"
    class PixelTexture(X3DElement): tag = "PixelTexture"
    class PixelTexture3D(X3DElement): tag = "PixelTexture3D"
    class Plane(X3DElement): tag = "Plane"
    class PlaneSensor(X3DElement): tag = "PlaneSensor"
    class PointLight(X3DElement): tag = "PointLight"
    class PointProperties(X3DElement): tag = "PointProperties"
    class PointSet(X3DElement): tag = "PointSet"
    class Polyline2D(X3DElement): tag = "Polyline2D"
    class Polypoint2D(X3DElement): tag = "Polypoint2D"
    class PopGeometry(X3DElement): tag = "PopGeometry"
    class PopGeometryLevel(X3DElement): tag = "PopGeometryLevel"
    class PositionChaser(X3DElement): tag = "PositionChaser"
    class PositionChaser2D(X3DElement): tag = "PositionChaser2D"
    class PositionDamper(X3DElement): tag = "PositionDamper"
    class PositionDamper2D(X3DElement): tag = "PositionDamper2D"
    class PositionInterpolator(X3DElement): tag = "PositionInterpolator"
    class PositionInterpolator2D(X3DElement): tag = "PositionInterpolator2D"
    class ProjectionVolumeStyle(X3DElement): tag = "ProjectionVolumeStyle"
    class Pyramid(X3DElement): tag = "Pyramid"
    class QuadSet(X3DElement): tag = "QuadSet"
    class RadarVolumeStyle(X3DElement): tag = "RadarVolumeStyle"
    class Rectangle2D(X3DElement): tag = "Rectangle2D"
    class RectangularTorus(X3DElement): tag = "RectangularTorus"
    class RefinementTexture(X3DElement): tag = "RefinementTexture"
    class RemoteSelectionGroup(X3DElement): tag = "RemoteSelectionGroup"
    class RenderedTexture(X3DElement): tag = "RenderedTexture"
    class RigidBody(X3DElement): tag = "RigidBody"
    class RigidBodyCollection(X3DElement): tag = "RigidBodyCollection"
    class Route(X3DElement): tag = "Route"
    class ScalarChaser(X3DElement): tag = "ScalarChaser"
    class ScalarDamper(X3DElement): tag = "ScalarDamper"
    class ScalarInterpolator(X3DElement): tag = "ScalarInterpolator"
    class Scene(X3DElement): tag = "Scene"
    class SegmentedVolumeData(X3DElement): tag = "SegmentedVolumeData"
    class ShadedVolumeStyle(X3DElement): tag = "ShadedVolumeStyle"
    class ShaderPart(X3DElement): tag = "ShaderPart"
    class Shape(X3DElement): tag = "Shape"
    class SilhouetteEnhancementVolumeStyle(X3DElement): tag = "SilhouetteEnhancementVolumeStyle"
    class SingleAxisHingeJoint(X3DElement): tag = "SingleAxisHingeJoint"
    class SliderJoint(X3DElement): tag = "SliderJoint"
    class SlopedCylinder(X3DElement): tag = "SlopedCylinder"
    class Snout(X3DElement): tag = "Snout"
    class SolidOfRevolution(X3DElement): tag = "SolidOfRevolution"
    class Sound(X3DElement): tag = "Sound"
    class Sphere(X3DElement): tag = "Sphere"
    class SphereSegment(X3DElement): tag = "SphereSegment"
    class SphereSensor(X3DElement): tag = "SphereSensor"
    class SplinePositionInterpolator(X3DElement): tag = "SplinePositionInterpolator"
    class SpotLight(X3DElement): tag = "SpotLight"
    class StaticGroup(X3DElement): tag = "StaticGroup"
    class StippleVolumeStyle(X3DElement): tag = "StippleVolumeStyle"
    class SurfaceShaderTexture(X3DElement): tag = "SurfaceShaderTexture"
    class Switch(X3DElement): tag = "Switch"
    class TexCoordDamper2D(X3DElement): tag = "TexCoordDamper2D"
    class Text(X3DElement): tag = "Text"
    class Texture(X3DElement): tag = "Texture"
    class TextureCoordinate(X3DElement): tag = "TextureCoordinate"
    class TextureCoordinate3D(X3DElement): tag = "TextureCoordinate3D"
    class TextureCoordinateGenerator(X3DElement): tag = "TextureCoordinateGenerator"
    class TextureProperties(X3DElement): tag = "TextureProperties"
    class TextureTransform(X3DElement): tag = "TextureTransform"
    class TextureTransform3D(X3DElement): tag = "TextureTransform3D"
    class TextureTransformMatrix3D(X3DElement): tag = "TextureTransformMatrix3D"
    class TimeSensor(X3DElement): tag = "TimeSensor"
    class IntegerSequencer(X3DElement): tag = "IntegerSequencer"
    class ToneMappedVolumeStyle(X3DElement): tag = "ToneMappedVolumeStyle"
    class Torus(X3DElement): tag = "Torus"
    class TouchSensor(X3DElement): tag = "TouchSensor"
    class Transform(X3DElement): tag = "Transform"
    class TriangleSet(X3DElement): tag = "TriangleSet"
    class TriangleSet2D(X3DElement): tag = "TriangleSet2D"
    class TwoSidedMaterial(X3DElement): tag = "TwoSidedMaterial"
    class Uniform(X3DElement): tag = "Uniform"
    class UniversalJoint(X3DElement): tag = "UniversalJoint"
    class Viewfrustum(X3DElement): tag = "Viewfrustum"
    class Viewpoint(X3DElement): tag = "Viewpoint"
    class VolumeData(X3DElement): tag = "VolumeData"
    class WorldInfo(X3DElement): tag = "WorldInfo"
    class X3DAppearanceChildNode(X3DElement): tag = "X3DAppearanceChildNode"
    class X3DAppearanceNode(X3DElement): tag = "X3DAppearanceNode"
    class X3DBackgroundNode(X3DElement): tag = "X3DBackgroundNode"
    class X3DBinaryContainerGeometryNode(X3DElement): tag = "X3DBinaryContainerGeometryNode"
    class X3DBindableNode(X3DElement): tag = "X3DBindableNode"
    class X3DBoundedObject(X3DElement): tag = "X3DBoundedObject"
    class X3DChaserNode(X3DElement): tag = "X3DChaserNode"
    class X3DChildNode(X3DElement): tag = "X3DChildNode"
    class X3DColorNode(X3DElement): tag = "X3DColorNode"
    class X3DComposableVolumeRenderStyleNode(X3DElement): tag = "X3DComposableVolumeRenderStyleNode"
    class X3DComposedGeometryNode(X3DElement): tag = "X3DComposedGeometryNode"
    class X3DCoordinateNode(X3DElement): tag = "X3DCoordinateNode"
    class X3DDamperNode(X3DElement): tag = "X3DDamperNode"
    class X3DDragSensorNode(X3DElement): tag = "X3DDragSensorNode"
    class X3DEnvironmentNode(X3DElement): tag = "X3DEnvironmentNode"
    class X3DEnvironmentTextureNode(X3DElement): tag = "X3DEnvironmentTextureNode"
    class X3DFogNode(X3DElement): tag = "X3DFogNode"
    class X3DFollowerNode(X3DElement): tag = "X3DFollowerNode"
    class X3DFontStyleNode(X3DElement): tag = "X3DFontStyleNode"
    class X3DGeometricPropertyNode(X3DElement): tag = "X3DGeometricPropertyNode"
    class X3DGeometryNode(X3DElement): tag = "X3DGeometryNode"
    class X3DGroupingNode(X3DElement): tag = "X3DGroupingNode"
    class X3DInfoNode(X3DElement): tag = "X3DInfoNode"
    class X3DInterpolatorNode(X3DElement): tag = "X3DInterpolatorNode"
    class X3DLODNode(X3DElement): tag = "X3DLODNode"
    class X3DLightNode(X3DElement): tag = "X3DLightNode"
    class X3DMaterialNode(X3DElement): tag = "X3DMaterialNode"
    class X3DMetadataObject(X3DElement): tag = "X3DMetadataObject"
    class X3DNBodyCollidableNode(X3DElement): tag = "X3DNBodyCollidableNode"
    class X3DNavigationInfoNode(X3DElement): tag = "X3DNavigationInfoNode"
    class X3DNode(X3DElement): tag = "X3DNode"
    class X3DPlanarGeometryNode(X3DElement): tag = "X3DPlanarGeometryNode"
    class X3DPointingDeviceSensorNode(X3DElement): tag = "X3DPointingDeviceSensorNode"
    class X3DRigidJointNode(X3DElement): tag = "X3DRigidJointNode"
    class X3DSensorNode(X3DElement): tag = "X3DSensorNode"
    class X3DShaderNode(X3DElement): tag = "X3DShaderNode"
    class X3DShapeNode(X3DElement): tag = "X3DShapeNode"
    class X3DSoundNode(X3DElement): tag = "X3DSoundNode"
    class X3DSoundSourceNode(X3DElement): tag = "X3DSoundSourceNode"
    class X3DSpatialGeometryNode(X3DElement): tag = "X3DSpatialGeometryNode"
    class X3DTexture3DNode(X3DElement): tag = "X3DTexture3DNode"
    class X3DTextureCoordinateNode(X3DElement): tag = "X3DTextureCoordinateNode"
    class X3DTextureNode(X3DElement): tag = "X3DTextureNode"
    class X3DTextureTransformNode(X3DElement): tag = "X3DTextureTransformNode"
    class X3DTimeDependentNode(X3DElement): tag = "X3DTimeDependentNode"
    class X3DTouchSensorNode(X3DElement): tag = "X3DTouchSensorNode"
    class X3DTransformNode(X3DElement): tag = "X3DTransformNode"
    class X3DVertexAttributeNode(X3DElement): tag = "X3DVertexAttributeNode"
    class X3DViewpointNode(X3DElement): tag = "X3DViewpointNode"
    class X3DVolumeDataNode(X3DElement): tag = "X3DVolumeDataNode"
    class X3DVolumeRenderStyleNode(X3DElement): tag = "X3DVolumeRenderStyleNode"




"""
Sets up a general Interpolator class that looks like Mathematica's InterpolatingFunction class
"""

import numpy as np, math, time, abc
import scipy.spatial as spat
import itertools, typing, dataclasses

import scipy.special as special

from ..Numputils import vec_outer, vec_tensordot, unique as nput_unique, vec_tensordiag, tensordot_deriv, tensorprod_deriv
from .Taylor import FunctionExpansion
from .Symbolic import TensorExpression
from ..Scaffolding import Logger

__all__ = [
    "RBFDInterpolator",
    "InverseDistanceWeightedInterpolator"
]

class NeighborBasedInterpolator(metaclass=abc.ABCMeta):
    """
    Useful base class for neighbor-based interpolation
    """

    __props__ = (
        "clustering_radius",
        "neighborhood_size",
        "neighborhood_merge_threshold",
        "neighborhood_clustering_radius",
        "coordinate_transform",
        "bad_interpolation_retries",
        "logger"
    )
    def __init__(self,
                 pts, values, *derivatives,
                 clustering_radius=None,
                 neighborhood_size=15,
                 neighborhood_merge_threshold=None,
                 neighborhood_max_merge_size=100,
                 neighborhood_clustering_radius=None,
                 coordinate_transform=None,
                 bad_interpolation_retries=2,
                 # max_interpolant_weight=1e25,
                 logger=None
                 ):

        pts = np.asanyarray(pts)
        values = np.asarray(values)
        derivatives = [np.asanyarray(d) for d in derivatives]
        # self._coord_shape = pts.shape[-1]
        if pts.ndim == 1:
            pts = pts[:, np.newaxis]
        elif pts.ndim > 2:
            pts = pts.reshape((pts.shape[0], -1))
            ders = []
            for i, d in enumerate(derivatives):
                ders.append(d.reshape((pts.shape[0],) + pts.shape[1:]*(i+1)))
            derivatives = ders
        # if clustering_radius is None:
        #     clustering_radius = .01**(1/pts.shape[-1])
        if clustering_radius is not None and clustering_radius >= 0:
            pts, values, derivatives = self.decluster_data(pts, values, derivatives, clustering_radius)
        self.grid = pts
        self.coordinate_transform=coordinate_transform

        self.tree = spat.KDTree(self.grid)
        self.neighborhood_size = neighborhood_size
        self.neighborhood_clustering_radius = neighborhood_clustering_radius
        self.merge_limit = neighborhood_merge_threshold
        self.max_merge_size = neighborhood_max_merge_size

        self.bad_interpolation_retries = bad_interpolation_retries

        self._pos_cache = {}
        self.vals = values
        self.derivs = derivatives

        self._globint = None

        self.logger = Logger.lookup(logger)

    @staticmethod
    def _decluster(pts, radius):
        mask = np.ones(len(pts), dtype=bool)
        for i in range(len(pts) - 1):
            if mask[i]:
                test_pos = np.where(mask[i + 1:])
                if len(test_pos) == 0 or len(test_pos[0]) == 0:
                    break
                samp_pos = i + 1 + test_pos[0]
                dists = np.linalg.norm(pts[samp_pos] - pts[i][np.newaxis], axis=1)
                mask[samp_pos] = dists > radius
        return pts[mask]
    @classmethod
    def decluster_data(cls, pts, vals, derivs, radius, return_mask=False): #TODO: I think this is broken maybe?
        rpts, _, _ = cls.RescalingData.renormalize_grid(pts)
        chunk_size = int(1e4)
        num_chunks = int(len(rpts)/chunk_size) + 1
        N = len(rpts)
        bad_pos = []
        for i in range(num_chunks):
            row_start = i * chunk_size
            row_end = min((i+1) * chunk_size, N)
            if row_start >= N:
                continue
            for j in range(num_chunks):
                col_start = j * chunk_size
                col_end = min((j+1) * chunk_size, N)
                if col_start >= N:
                    continue
                rinds, cinds = np.triu_indices(row_end - row_start, m=col_end - col_start, k=1)
                v = rpts[row_start + rinds] - rpts[col_start + cinds]
                dvec = np.linalg.norm(v, axis=1)
                bad_spots = dvec < radius
                if bad_spots.any():
                    col_start = j*chunk_size
                    bad_pos.append(col_start + cinds[bad_spots])

        if len(bad_pos) == 0 or len(bad_pos[0]) == 0:
            if return_mask:
                return pts, vals, derivs, np.ones(len(pts), dtype=bool)
            else:
                return pts, vals, derivs

        mask = np.ones(len(pts), dtype=bool)
        for b in bad_pos:
            mask[b] = False
        pts = pts[mask]
        vals = vals[mask]
        derivs = [d[mask] for d in derivs]

        if return_mask:
            return pts, vals, derivs, mask
        else:
            return pts, vals, derivs

    @dataclasses.dataclass
    class RescalingData:
        grid_shifts:'Iterable[float]'
        grid_scaling:'Iterable[float]'
        vals_shift:'float'
        vals_scaling:'float'
        __slots__ = [
            'grid_shifts', 'grid_scaling',
            'vals_shift', 'vals_scaling'
        ]

        @classmethod
        def initialize_subgrid_data(cls, pts, values, derivatives):
            grid, grid_shifts, grid_scaling = cls.renormalize_grid(pts)
            vals, val_shift, val_scaling = cls.renormalize_values(values)
            derivs = cls.renormalize_derivs(derivatives, val_scaling, grid_scaling)
            return grid, vals, derivs, cls(grid_shifts, grid_scaling, val_shift, val_scaling)

        @classmethod
        def renormalize_grid(cls, pts):
            dim_slices = np.transpose(pts)
            mins = [np.min(x) for x in dim_slices]
            maxs = [np.max(x) for x in dim_slices]
            scalings = [1 if M - m < 1e-8 else M - m for M, m in zip(maxs, mins)]
            rescaled_slices = [
                (x - m) / s for x, m, s in zip(dim_slices, mins, scalings)
            ]
            return np.transpose(rescaled_slices), np.array(mins), np.array(scalings)
        @classmethod
        def renormalize_values(cls, values):
            min = np.min(values)
            scaling = np.max(values) - min
            return (values - min) / scaling, min, scaling
        @classmethod
        def renormalize_derivs(cls, derivs, vals_scaling, grid_scaling):
            ders = []
            for n, values in enumerate(derivs):
                values = values / vals_scaling
                ndim = len(grid_scaling)
                gs = grid_scaling
                for d in range(n):
                    gs = np.expand_dims(gs, 0)
                    grid_scaling = grid_scaling[..., np.newaxis] * gs
                values = values * grid_scaling
                inds = (slice(None, None, None),) + tuple(
                    np.array(list(itertools.combinations_with_replacement(range(ndim), r=n + 1))).T
                )
                ders.append(values[inds])
            return ders

        def apply_renormalization(self, pts):
            pts = pts.T
            scaligns = self.grid_scaling.copy()
            scaligns[scaligns == 0] = 1. # only gonna happen when there's no shift anyway...
            return ((pts - self.grid_shifts[:, np.newaxis]) / scaligns[:, np.newaxis]).T
        def reverse_renormalization(self, pts):
            return pts * self.grid_scaling + self.grid_shifts

        def reverse_value_renormalization(self, vs):
            return vs * self.vals_scaling + self.vals_shift

        def reverse_derivative_renormalization(self, derivs, reshape=True):
            ders = []
            npts = None
            ndim = len(self.grid_scaling)
            for n, values in enumerate(derivs):

                # we assume we have npts x (ndim,)*n so we reshape
                if npts is None: # we start with 1D
                    npts = len(values) / ndim
                    npp = int(npts)
                    if npts != npp:
                        raise ValueError("wtf")
                    npts = npp

                k = n+1

                shp = (npts,) + (ndim,) * k
                if reshape:
                    if np.prod(shp) == len(values):
                        values = values.reshape(shp)
                    else:
                        a = np.zeros(shp)
                        base_inds = NeighborBasedInterpolator.triu_inds(ndim, k)
                        inds = (slice(None, None, None),) + base_inds
                        a[inds] = values.reshape((npts, -1))
                        for pos in np.array(base_inds).T:
                            base = (slice(None, None, None),) + tuple(pos)
                            for p in itertools.permutations(pos):
                                ind = (slice(None, None, None),) + p
                                a[ind] = a[base]
                        values = a
                else:
                    values = values.reshape((npts, -1))

                values = values * self.vals_scaling
                grid_scaling = self.grid_scaling
                gs = grid_scaling
                for d in range(n): # for a small performance boost I can avoid doing this repeatedly
                    gs = np.expand_dims(gs, 0)
                    grid_scaling = grid_scaling[..., np.newaxis] * gs
                if reshape:
                    values = values / grid_scaling
                else:
                    inds = RBFDInterpolator.triu_inds(ndim, k)
                    values = values / grid_scaling[inds]
                # inds = (slice(None, None, None),) + tuple(
                #     np.array(list(itertools.combinations_with_replacement(range(ndim), r=n + 1))).T
                # )
                ders.append(values)
            return ders

    @staticmethod # this is surprisingly slow and totally cacheable
    def triu_inds(ndim, rank):
        return tuple(
            np.array(
                list(itertools.combinations_with_replacement(range(ndim), r=rank))
            ).T
        )
    @staticmethod
    def triu_num(ndim, rank):
        if ndim == 1:
            return 1
        else:
            return math.comb(rank+(ndim-1), (ndim-1))

    def get_neighborhood(self, pts, *, neighbors, return_distances=False):
        if neighbors is None:
            neighbors = self.neighborhood_size
        distances, yindices = self.tree.query(pts, k=neighbors)
        if neighbors == 1:
            # `KDTree` squeezes the output when neighbors=1.
            yindices = yindices[:, np.newaxis]
            distances = distances[:, np.newaxis]
        # if self.grid.shape[-1] == 3:
        #     for p,idx,ddd in zip(pts, yindices, distances):
        #         nss = self.grid[idx]
        #         print(f"{p} --> {nss} ({ddd})")

        res = yindices
        if return_distances:
            res = (yindices, distances)
        return res

    def create_neighbor_groups(self, inds, merge_limit=None, max_merge_size=None):
        #TODO: lots of optimizations here, cutting down on loops, making use of sorting
        #      not doing useless setdiffs, etc.
        # _, sorting, subinds = nput_unique(inds, axis=0, return_index=True)
        if merge_limit is None:
            merge_limit = self.merge_limit
        if max_merge_size is None:
            max_merge_size = self.max_merge_size

        inds, sorting, counts = nput_unique(inds, axis=0, return_counts=True)
        subinds = np.split(sorting, np.cumsum(counts))[:-1]
        # raise Exception(nput_unique(inds, axis=0, return_index=True))
        if merge_limit is not None:
            allowed_diffs = len(inds[0]) - merge_limit
        else:
            allowed_diffs = -1
        if allowed_diffs > 0:
            groups = []
            idx_mapping = []
            for i,idx in zip(subinds, inds):
                for n,g in enumerate(groups):
                    if len(g) < self.max_merge_size:
                        idx = idx[idx > 0]
                        g = g[g > 0]
                        diff = np.setdiff1d(idx, g, assume_unique=True)
                        if len(diff) < allowed_diffs:
                            groups[n] = np.sort(np.concatenate([g, diff]))
                            idx_mapping[n] = np.concatenate([idx_mapping[n], i])
                            break
                else:
                    groups.append(idx[idx > 0])
                    idx_mapping.append(i)
        else:
            groups = inds
            idx_mapping = subinds
        # print(np.cumsum(counts),
        #       idx_mapping)

        return groups, idx_mapping

    def prep_interpolation_data(self, inds):
        grid = self.grid[inds]
        vals = self.vals[inds]
        derivs = [d[inds] for d in self.derivs]

        # print("=>", grid.shape, vals.shape, [d.shape for d in derivs])

        grid, vals, derivs, scaling_data = self.RescalingData.initialize_subgrid_data(
            grid,
            vals,
            derivs
        )

        return grid, vals, derivs, scaling_data

    @abc.abstractmethod
    def construct_interpolation(self, inds, solver_data=False, return_error=False):
        ...

    @abc.abstractmethod
    def apply_interpolation(self, pts, data, inds, deriv_order=0,
                            reshape_derivatives=True,
                            return_data=False
                            ):
        """

        :param pts:
        :type pts:
        :param data:
        :type data:
        :param deriv_order:
        :type deriv_order:
        :return:
        :rtype:
        """

    def prep_neighborhoods(self,
                           pts,
                           hoods,
                           distances,
                           neighbors,
                           merge_neighbors=None,
                           neighborhood_clustering_radius=None,
                           min_distance=None,
                           max_distance=None,
                           use_natural_neighbors=None
                           ):

        if neighborhood_clustering_radius is None:
            neighborhood_clustering_radius = self.neighborhood_clustering_radius
        if neighborhood_clustering_radius is not None:
            self.logger.log_print("declustering neighbor data...", log_level=Logger.LogLevel.Debug)
            subhoods = self.grid[hoods]
            row_inds, col_inds = np.triu_indices(neighbors, k=1)
            subdists = np.linalg.norm(
                subhoods[:, row_inds] - subhoods[:, col_inds],
                axis=-1
            )
            mask = subdists < neighborhood_clustering_radius
            for i in range(len(mask)):
                kill_me = np.where(mask[i])
                if len(kill_me) > 0 and len(kill_me[0]) > 0:
                    hoods[i][col_inds[kill_me[0]]] = -1

        if min_distance is not None or max_distance is not None:
            dist_sets = distances.copy()
            ldr = np.arange(len(dist_sets), dtype=np.int)
            nds = np.argmin(dist_sets, axis=1)
            hood_nearest = hoods[ldr, nds]
            # dist_vals = dists[ldr, nds]
            nds2 = np.argmax(dist_sets, axis=1)
            hood_furthest = hoods[ldr, nds2]
            # dists[ldr, nds] = dist_vals # reset this for later

            if min_distance is not None:
                self.logger.log_print("applying min-distance cutoff...", log_level=Logger.LogLevel.Debug)
                dist_mask = dist_sets < min_distance

                # we're gonna drop points below the distance cutoff, then
                # resample the distances so we have close the the target number of
                # neighbors
                target_neighbors = int(neighbors / 2)

                to_shrink = np.sum(dist_mask,
                                   axis=1) < target_neighbors  # we dropped less than half of the extra neighbors
                if np.any(to_shrink):
                    # now we find the first non-dropped position for these chunks
                    dist_sets[dist_mask] = np.max(dist_sets)  # this will be -1 later

                    start_pos = np.argmin(dist_sets[to_shrink], axis=1)
                    end_pos = start_pos + target_neighbors
                    subhoods = hoods[to_shrink]
                    for i, e in enumerate(end_pos):  # drop all excess points
                        subhoods[e:] = -1
                    dist_sets[dist_mask] = -1  # so we can later apply a max cutoff

                submask = np.where(np.all(dist_mask, axis=1))  # make sure we're not dropping everything...
                if len(submask) > 0:  # we'll leave in the two nearest points since we're just extrapolating basically
                    set_mask = submask + (np.zeros(len(submask[0]), dtype=np.uint8),)
                    hoods[set_mask] = hood_nearest[submask]

            if max_distance is not None:
                self.logger.log_print("applying max-distance cutoff...", log_level=Logger.LogLevel.Debug)
                dist_mask = dist_sets > max_distance
                hoods[dist_mask] = -1
                submask = np.where(np.all(dist_mask, axis=1))
                if len(submask) > 0:
                    set_mask = submask + (np.zeros(len(submask[0]), dtype=np.uint8),)
                    hoods[set_mask] = hood_nearest[submask]
                    set_mask = submask + (np.ones(len(submask[0]), dtype=np.uint8),)
                    hoods[set_mask] = hood_furthest[submask]

        if use_natural_neighbors:
            self.logger.log_print("getting natural neighbors...", log_level=Logger.LogLevel.Debug)
            for p, hh in zip(pts, hoods):
                h = hh[hh > -1]
                if len(h) > 2 * pts.shape[-1]:
                    # shit_pts = self.grid[h]
                    # shit_rows, shit_cols = np.triu_indices(len(shit_pts), k=1)
                    # raise Exception(
                    #     np.min(
                    #         np.linalg.norm(shit_pts[shit_rows] - shit_pts[shit_cols], axis=-1)
                    #     )
                    # )
                    # raise Exception(
                    #     [(np.min(self.grid[h][:, i]), np.max(self.grid[h][i])) for i in range(self.grid[h].shape[-1])]
                    # )
                    # raise Exception(self.grid[h].shape)
                    mesh_pts = np.concatenate([self.grid[h], [p]], axis=0)
                    triang = spat.Delaunay(mesh_pts, qhull_options='QJ')
                    indptr, indices = triang.vertex_neighbor_vertices
                    hood = indices[indptr[len(mesh_pts) - 1]:indptr[len(mesh_pts)]]
                    # print(len(h), len(hood))
                    hh[:len(hood)] = h[hood]
                    hh[len(hood):] = -1
                # raise Exception(hood)
            # raise Exception(hoods)
        if len(hoods) > 0:
            ind_grps, pts_grps = self.create_neighbor_groups(hoods, merge_limit=merge_neighbors)
        else:
            ind_grps = pts_grps = []

        return ind_grps, pts_grps

    def eval(self,
             pts,
             deriv_order=0,
             neighbors=None,
             merge_neighbors=None,
             reshape_derivatives=True,
             return_interpolation_data=False,
             check_in_sample=True,
             zero_tol=1e-8,
             return_error=False,
             use_cache=True,
             retries=None,
             # resiliance_test_options=None,
             # extrapolation_warning=None
             max_distance=None,
             min_distance=None,
             neighborhood_clustering_radius=None,
             use_natural_neighbors=False,
             chunk_size=None
             ):

        pts = np.asanyarray(pts)
        # check against coord shape ?
        # if pts.shape[-len(self._coord_shape):] == self._coord_shape:
        #     pts.reshape()
        if pts.ndim == 1:
            pts = pts[np.newaxis]

        if pts.ndim > 2:
            extra_shape = pts.shape[:-1]
            pts = pts.reshape(-1, pts.shape[-1])
        else:
            extra_shape = None

        if retries is None:
            retries = self.bad_interpolation_retries
        with self.logger.block(
                tag="evaluating interpolation over {n} points with {o}-order derivatives".format(n=len(pts),
                                                                                                 o=deriv_order),
                log_level=Logger.LogLevel.Debug
        ):
            # set up return locations...
            val_sets = np.empty(len(pts))
            interp_data = np.full(len(pts), None) if return_interpolation_data else None
            der_sets = [
                np.empty((len(pts),) + d.shape[1:])
                    if reshape_derivatives else
                np.empty((len(pts), self.triu_num(pts.shape[-1], d.ndim - 1)))
                for d in self.derivs
            ] if deriv_order > 0 else None
            errors = np.empty((len(pts), 2)) if return_error else None

            if neighbors is None:
                neighbors = self.neighborhood_size
            if min_distance is not None:
                neighbors = 2 * neighbors

            if chunk_size is None:
                pts_chunks = [pts]
            else:
                num_chunk = int(len(pts) / chunk_size) + 1
                pts_chunks = np.array_split(pts, num_chunk)
            chunk_offset = 0

            cache = self._pos_cache if use_cache else {}

            for k, pts in enumerate(pts_chunks):

                self.logger.log_print("applying interpolation for chunk {k} (size={s})...", k=k, s=len(pts),
                                      log_level=Logger.LogLevel.Debug)
                hoods, dists = self.get_neighborhood(pts, neighbors=neighbors,
                                                     return_distances=check_in_sample or min_distance is not None)

                sub_map = None  # to map from the reduced sample to the full one if we are able to reduce
                if check_in_sample:  # look for things very close to an abcissa
                    zero_pos = np.where(dists < zero_tol)
                    if len(zero_pos) > 0:
                        sample_pos = hoods[zero_pos]
                        zero_pos = zero_pos[0]
                        if len(zero_pos) > 0:
                            val_sets[zero_pos] = self.vals[sample_pos]
                            if deriv_order > 0:
                                for n, d in enumerate(self.derivs):
                                    der_sets[n][zero_pos] = d[sample_pos]
                            if return_error:
                                errors[zero_pos] = np.broadcast_to([[0, 0]], (len(zero_pos), 2))
                            sub_map = np.setdiff1d(np.arange(len(pts)), zero_pos)
                            hoods = hoods[sub_map]
                            pts = pts[sub_map]
                            dists = dists[sub_map]

                ind_grps, pts_grps = self.prep_neighborhoods(
                    pts,
                    hoods,
                    dists,
                    neighbors,
                    merge_neighbors=merge_neighbors,
                    neighborhood_clustering_radius=neighborhood_clustering_radius,
                    min_distance=min_distance,
                    max_distance=max_distance,
                    use_natural_neighbors=use_natural_neighbors
                )

                if len(ind_grps) > 0:
                    self.logger.log_print("solving and applying interpolation equations for {n} neighborhoods",
                                          n=len(pts_grps), log_level=Logger.LogLevel.Debug)
                    self.logger.log_print("largest neighborhood: {N}", N=max(len(x) for x in ind_grps),
                                          log_level=Logger.LogLevel.Debug)

                print_block_modulus = 25  # 1 if pts.shape[-1] > 5 else 25
                block_start_time = None
                for gn, (pts_inds, vals_inds, ds) in enumerate(zip(pts_grps, ind_grps, dists)):  # we need to evaluate this entire loop for each point......
                    # ds = ds[vals_inds > -1]
                    vals_inds = vals_inds[vals_inds > -1]
                    if gn % print_block_modulus == 0:
                        if block_start_time is not None:
                            self.logger.log_print("    block took {s:.3f}s", s=time.time() - block_start_time,
                                                  log_level=Logger.LogLevel.Debug
                                                  )
                        block_start_time = time.time()
                        self.logger.log_print("  starting group {gn} (doing {gn} to {gnm})", gn=gn, size=len(vals_inds),
                                              gnm=gn + print_block_modulus - 1,
                                              log_level=Logger.LogLevel.Debug
                                              )

                        self.logger.log_print("  group sizes: {gss}",
                                              gss=[np.sum(ind_grps[gg] > 0) for gg in
                                                   range(gn, min(len(ind_grps), gn + print_block_modulus))],
                                              log_level=Logger.LogLevel.Debug
                                              )
                    # if len(pts_inds) != len(vals_inds):
                    #     raise ValueError(f"Mismatch between points {pts_inds} and values {vals_inds}")

                    # print(vals_inds)
                    # print("dist data:", np.min(ds), np.max(ds), np.std(ds))
                    k = tuple(vals_inds)
                    if len(k) == 1:
                        res = (self.vals[k], [d[k] for d in self.derivs])
                        if return_interpolation_data:
                            res = res + (None,)
                    else:
                        if k not in cache:
                            neighs = self.neighborhood_size if neighbors is None else neighbors
                            et = self.error_threshold  # for the cases where we have errors _at the points_
                            retries = max(retries, 1)
                            for n in range(retries):
                                try:
                                    if n == retries - 1:
                                        self.error_threshold = None  # always gotta succeed in the end...
                                    interp_stuff = self.construct_interpolation(vals_inds,
                                                                                solver_data=return_interpolation_data,
                                                                                return_error=return_error
                                                                                )
                                except RBFDError:
                                    # print(">>", n, vals_inds)
                                    neighs = neighs * 2  # just try to expand a little bit to fix things...
                                    new_hoods, new_dists = self.get_neighborhood(pts[pts_inds], neighbors=neighs, return_distances=True)
                                    # vals_inds = np.unique(np.concatenate([vals_inds, new_nearest.flatten()]))

                                    prepped_inds, _ = self.prep_neighborhoods(
                                        pts[pts_inds],
                                        new_hoods,
                                        new_dists,
                                        neighbors,
                                        merge_neighbors=-1,
                                        neighborhood_clustering_radius=neighborhood_clustering_radius,
                                        min_distance=min_distance,
                                        max_distance=max_distance,
                                        use_natural_neighbors=use_natural_neighbors
                                    )
                                    vals_inds = np.concatenate([p[p>-1] for p in prepped_inds])

                                else:
                                    # if resiliance_test_options is not None:
                                    #     expansion = self.construct_function_expansion(vals_inds)
                                    #     self.resiliance_test(expansion, interp_stuff, **resiliance_test_options)
                                    break
                                finally:
                                    self.error_threshold = et
                            if return_error:
                                errors[pts_inds, :] = np.broadcast_to(interp_stuff.interpolation_error,
                                                                      (len(pts_inds), 2))
                            cache[k] = interp_stuff
                        res = self.apply_interpolation(pts[pts_inds], cache[k], vals_inds,
                                                       deriv_order=deriv_order,
                                                       reshape_derivatives=reshape_derivatives,
                                                       return_data=return_interpolation_data
                                                       )
                    res_inds = chunk_offset + (pts_inds if sub_map is None else sub_map[pts_inds])
                    if return_interpolation_data:
                        vals, ders, mat = res
                        interp_data[res_inds] = {
                            'points': self.grid[vals_inds],
                            'data': cache[k],
                            'matrix': mat,
                            'solver_data': cache[k].solver_data
                        }
                    else:
                        vals, ders = res
                    val_sets[res_inds] = vals
                    if deriv_order > 0:
                        for n, d in enumerate(ders):
                            der_sets[n][res_inds] = d

                    # print(vals)
                    # if deriv_order > 0:
                    #     print(ders)

                else:
                    if block_start_time is not None:
                        self.logger.log_print("    final block took {s:.3f}s", s=time.time() - block_start_time,
                                              log_level=Logger.LogLevel.Debug
                                              )

                chunk_offset += len(pts)

            self.logger.log_print("reshaping interpolated data", log_level=Logger.LogLevel.Debug)

            if extra_shape is not None:
                val_sets = np.reshape(val_sets, extra_shape)
                if return_interpolation_data:
                    interp_data = np.reshape(interp_data, extra_shape)
                if return_error:
                    errors = np.reshape(errors, extra_shape + (2,))
                if der_sets is not None:
                    der_sets = [
                        d.reshape(extra_shape + d.shape[1:])
                        for d in der_sets
                    ]

            if deriv_order > 0:
                if der_sets is None and len(pts) == 0:
                    der_sets = []
                res = [val_sets] + der_sets
            else:
                res = val_sets

            if return_interpolation_data or return_error:
                res = (res,)

            if return_error:
                res = res + (errors,)

            if return_interpolation_data:
                res = res + (interp_data,)

        return res

    def resiliance_test(self, expansion, interpolation_data, mesh_spacing=0.01, tolerance=0.05):
        raise NotImplementedError("changed things up...")
        # allow 5% error (maybe scaled by mesh spacing)?
        # -> presumably we have some kind of assurance based on the
        # order of the polynomial approximation for how large our error is?
        # we could calculate the interpolation at all of its target points...?

        points = expansion.center + mesh_spacing
        expand = expansion(points, outer=False)
        interp = self.apply_interpolation(points, interpolation_data)[0]
        m = np.min([np.min(expand), np.min(interp)])
        M = np.max([np.max(expand), np.max(interp)])

        error = np.abs(expand - interp) / (M - m)

        if np.max(error) > tolerance:
            raise RBFDError(
                "failed resiliance test (mesh spacing {} leads to scaled deviation from expansion greater than {})".format(
                    mesh_spacing,
                    tolerance
                ))

    def __call__(self, pts, deriv_order=0,
                 neighbors=None,
                 merge_neighbors=None,
                 reshape_derivatives=True,
                 return_interpolation_data=False,
                 use_cache=True,
                 return_error=False,
                 zero_tol=1e-8,
                 retries=None,
                 **extra_opts
                 ):
        return self.eval(pts,
                         deriv_order=deriv_order, neighbors=neighbors,
                         merge_neighbors=merge_neighbors, reshape_derivatives=reshape_derivatives,
                         return_interpolation_data=return_interpolation_data,
                         use_cache=use_cache,
                         return_error=return_error,
                         zero_tol=zero_tol,
                         retries=retries,
                         **extra_opts
                         )


    def construct_function_expansion(self, inds):
        grid = self.grid[inds]
        vals = self.vals[inds]
        derivs = [d[inds] for d in self.derivs]

        return FunctionExpansion(
            derivs,
            center=grid,
            ref=vals,
            weight_coefficients=False # I'm _pretty_ sure that's what I want?
        )

    @classmethod
    def create_function_interpolation(cls,
                                      pts,
                                      fn,
                                      *derivatives,
                                      derivative_order=None,
                                      function_shape=None,
                                      **opts):
        from ..Scaffolding import ParameterManager
        from .Taylor import FiniteDifferenceDerivative

        if derivative_order is None:
            derivative_order = len(derivatives)
        vals = fn(pts)
        dvals = []
        for n,d in enumerate(derivatives):
            if n < derivative_order:
                dvals.append(d(pts))
            else:
                break

        if pts.ndim == 1:
            input_shape = 0
        else:
            input_shape = pts.shape[-1]

        opts = ParameterManager(opts)
        if len(dvals) < derivative_order:
            fd_opts = opts.filter(FiniteDifferenceDerivative)
            base_deriv_order = len(derivatives)
            fdd = FiniteDifferenceDerivative(
                derivatives[-1] if base_deriv_order > 0 else fn,
                function_shape=(input_shape, 0) if function_shape is None else function_shape
            )

            deriv_tensors = fdd.derivatives(
                center=pts,
                **fd_opts
            ).compute_derivatives(list(range(1, derivative_order - base_deriv_order + 1)))

            dvals.extend(
                np.moveaxis(a, 0, -1) for a in deriv_tensors[base_deriv_order:]
            )

        return cls(
            pts,
            vals,
            *dvals,
            **opts.exclude(FiniteDifferenceDerivative)
        )

    def nearest_interpolation(self, pts, neighbors=None, solver_data=False, interpolator=True):
        """

        :param pts:
        :type pts:
        :param neighbors:
        :type neighbors:
        :param solver_data:
        :type solver_data:
        :param interpolator:
        :type interpolator:
        :return:
        :rtype: RBFDInterpolator.Interpolator|RBFDInterpolator.InterpolationData
        """
        pts = np.asanyarray(pts)
        smol = pts.ndim == 1
        if smol:
            pts = pts[np.newaxis]
        inds = self.get_neighborhood(pts, neighbors=neighbors)
        dats = [self.construct_interpolation(i, solver_data=solver_data) for i in inds]
        if interpolator:
            dats = [self.Interpolator(d, self) for d in dats]
        if smol:
            dats = dats[0]
        return dats

    class Interpolator:
        __slots__ = ['data', 'inds', 'parent']
        def __init__(self, data, inds, parent:'NeighborBasedInterpolator'):
            self.data = data
            self.parent = parent
            self.inds = inds
        def __call__(self, pts, deriv_order=0, reshape_derivatives=True):
            pts = np.asanyarray(pts)
            smol = pts.ndim == 1
            if smol:
                pts = pts[np.newaxis]
            vals, ders = self.parent.apply_interpolation(pts, self.data, self.inds,
                                                   deriv_order=deriv_order,
                                                   reshape_derivatives=reshape_derivatives
                                                   )
            # TODO: ADD ANY RESHAPING
            if deriv_order > 0:
                return [vals] + ders
            else:
                return vals


    @property
    def global_interpolator(self):
        if self._globint is None:
            et = self.error_threshold
            try: # ignore errors
                self.error_threshold = None
                interp = self.construct_interpolation(np.arange(len(self.grid)))
            finally:
                self.error_threshold = et
            self._globint = self.Interpolator(interp, np.arange(len(self.grid)), self)
        return self._globint

class RBFDError(ValueError):
    ...

class RBFDInterpolator(NeighborBasedInterpolator):
    """
    Provides a flexible RBF interpolator that also allows
    for matching function derivatives
    """

    def __init__(self,
                 pts, values, *derivatives,
                 kernel:typing.Union[callable,dict]='thin_plate_spline',
                 kernel_options=None,
                 auxiliary_basis=None,
                 auxiliary_basis_options=None,
                 extra_degree=0,
                 clustering_radius=None,
                 monomial_basis=True,
                 multicenter_monomials=True,
                 neighborhood_size=15,
                 neighborhood_merge_threshold=None,
                 neighborhood_max_merge_size=100,
                 neighborhood_clustering_radius=None,
                 solve_method='svd',
                 max_condition_number=np.inf,
                 error_threshold=.01,
                 bad_interpolation_retries=3,
                 coordinate_transform=None,
                 # max_interpolant_weight=1e25,
                 logger=None
                 ):

        super().__init__(pts, values, *derivatives,
                         neighborhood_size=neighborhood_size,
                         neighborhood_merge_threshold=neighborhood_merge_threshold,
                         neighborhood_max_merge_size=neighborhood_max_merge_size,
                         neighborhood_clustering_radius=neighborhood_clustering_radius,
                         clustering_radius=clustering_radius,
                         bad_interpolation_retries=bad_interpolation_retries,
                         coordinate_transform=coordinate_transform,
                         logger=logger
                         )

        self.solve_method = solve_method
        self.max_condition_number = max_condition_number # not actually using this...
        # self.max_interpolant_weight = max_interpolant_weight # hopefully this is enough...
        self.error_threshold = error_threshold

        if isinstance(kernel, str):
            kernel = self.default_kernels[kernel]
        if not isinstance(kernel, dict):
            kernel = {'function':kernel}
        if 'derivatives' not in kernel:
            kernel['derivatives'] = None
        if 'zero_handler' not in kernel:
            kernel['zero_handler'] = None
        if kernel_options is not None:
            kernel['function'] = lambda r,fn=kernel['function'],opt=kernel_options,inds=None:fn(r,inds=inds,**opt)
            if kernel['derivatives'] is not None:
                kernel['derivatives'] = lambda n, fn=kernel['derivatives'],opt=kernel_options,inds=None: (lambda r,d=fn(n):d(r, inds=inds, **opt))
            if kernel['zero_handler'] is not None:
                kernel['zero_handler'] = lambda n,ndim,fn=kernel['zero_handler'],opt=kernel_options,inds=None: (lambda r,d=fn(n, ndim):d(r, inds=inds, **opt))
        self.kernel = kernel

        if isinstance(auxiliary_basis, str):
            auxiliary_basis = self.default_auxiliary_bases[auxiliary_basis]
        if not isinstance(auxiliary_basis, dict):
            auxiliary_basis = {'function':auxiliary_basis}
        if 'derivatives' not in auxiliary_basis:
            auxiliary_basis['derivatives'] = None
        if auxiliary_basis_options is not None:
            auxiliary_basis['function'] = lambda r,fn=auxiliary_basis['function'],opt=auxiliary_basis_options,inds=None:fn(r,inds=inds,**opt)
            if auxiliary_basis['derivatives'] is not None:
                auxiliary_basis['derivatives'] = lambda n, fn=auxiliary_basis['derivatives'],opt=auxiliary_basis_options,inds=None: (lambda r,d=fn(n):d(r, inds=inds, **opt))
        self.aux_poly = auxiliary_basis
        self.extra_degree = extra_degree
        self.monomial_basis = monomial_basis
        self.multicenter_monomials = multicenter_monomials

        self._expr_cache = {}

    # def to_state(self, serializer=None):
    #     """
    #     Provides just the state that is needed to
    #     serialize the object
    #     :param serializer:
    #     :type serializer:
    #     :return:
    #     :rtype:
    #     """
    #     return {
    #         'grid': self.grid,
    #         'vals': self.vals,
    #         'derivs': self.derivs,
    #         'kernel': self.kernel,
    #         'neighborhood_size': self.neighborhood_size,
    #         'aux_poly': self.aux_poly,
    #         'monomial_basis': self.monomial_basis,
    #         'multicenter_monomials': self.multicenter_monomials,
    #         'merge_limit': self.merge_limit,
    #         'max_merge_size': self.max_merge_size,
    #         'extra_degree': self.extra_degree,
    #         'solve_method': self.solve_method,
    #         'max_condition_number': self.max_condition_number,
    #         'error_threshold': self.error_threshold,
    #     }
    # @classmethod
    # def from_state(cls, state, serializer=None):
    #     return cls(
    #         state['grid'],
    #         state['vals'],
    #         *state['derivs'],
    #         kernel=state['kernel'],
    #         neighborhood_size=state['neighborhood_size'],
    #         neighborhood_merge_threshold=state['merge_limit'],
    #         neighborhood_max_merge_size=state['max_merge_size'],
    #         solve_method=state['solve_method'],
    #         max_condition_number=state['max_condition_number'],
    #         error_threshold=state['error_threshold'],
    #         auxiliary_basis=state['aux_poly'],
    #         extra_degree=state['extra_degree'],
    #         monomial_basis=state['monomial_basis'],
    #         multicenter_monomials=state['multicenter_monomials'],
    #     )

    # TODO: add more structure defining RBF class and have each type of interpolant
    #       specialize that

    @staticmethod
    def gaussian(r, e=1, inds=None):
        if inds is not None and not isinstance(e, (int, float, np.integer, np.floating)):
            e = e[inds]
        return np.exp(-(r**2)/(2*e**2))
    @staticmethod
    def gaussian_derivative(n:int, inds=None):
        from scipy.special import hermite
        s = -1 if n%2 == 1 else 1
        def deriv(r, e=1, inds=inds):
            if inds is not None and not isinstance(e, (int, float, np.integer, np.floating)):
                e = e[inds]
            w = 1 / (e * np.sqrt(2))
            r = w * r
            return s*(w**n)*hermite(n)(r)*np.exp(-r**2)
        return deriv
    @staticmethod
    def gaussian_singularity_handler(n: int, ndim: int, inds=None):
        # handler for what happens at zero specifically
        def deriv(r, e=1, inds=inds): # r will be a vector of zeros..., inds=None):
            if inds is not None and not isinstance(e, (int, float, np.integer, np.floating)):
                e = e[inds]
            dmat = np.zeros((len(r),) + (ndim,)*n)
            if n % 2 == 0:
                crds = np.arange(ndim)
                # take tuples of crds, could be faster but we shouldn't
                # need many evals of this...
                for tup in itertools.combinations_with_replacement(crds, r=n//2):
                    idx, counts = np.unique(tup, return_counts=True)
                    val = np.prod([np.arange(1, 2*c+1, 2) for c in counts])/(e**n) # rising factorial
                    idx = sum(
                        [
                            [i] * (2 * c) for i, c in
                            zip(idx, counts)
                        ],
                        []
                    )
                    for p in itertools.permutations(idx):
                        p = (slice(None, None, None),) + p
                        dmat[p] = val
            return ((-1)**(n//2))*dmat
        return deriv

    @staticmethod
    def thin_plate_spline(r, o=3, inds=None):
        return r**o * (1 if o%2 == 1 else np.log(r+1))
    @staticmethod
    def thin_plate_spline_derivative(n: int, inds=None):
        def deriv(r, o=3, inds=inds):
            if o % 2 == 1:
                k = np.prod(o-np.arange(n))
                return k * r**(o-n)
            elif n == 1: # clearly could work this out but I don't want to right now
                """r^n/(1 + r) + n r^(-1 + n) Log[1 + r]"""
                return r**n/(1+r) + n*r**(n-1)*np.log(1+r)
            elif n == 2:
                """-(r^n/(1 + r)^2) + (2 n r^(-1 + n))/(1 + r) + (-1 + n) n r^(-2 + n) Log[1 + r]"""
                return -(r**n/(1+r)**2) + (2*n*r**(n-1))/(1+r) + n*(n-1)*r**(n-2)*np.log(1+r)
            else:
                raise NotImplementedError("need higher derivs for even thin plate splines...")
        return deriv
    @staticmethod
    def thin_plate_spline_singularity_handler(n: int, ndim: int, inds=None):
        # handler for what happens at zero specifically
        def deriv(r, o=3, inds=inds):  # r will be a vector of zeros...
            if n - (1 - (o%2)) > o:
                raise ValueError("thin plate spline deriv not defined")
            return  np.zeros((len(r),) + (ndim,) * n)
        return deriv

    wendland_coefficient_cache = {}
    @classmethod
    def wendland_coefficient(cls, l, j, k):
        if k == 0:
            return (-1)**j * special.binom(l, j)
        elif j == 1:
            return 0
        else:
            if (l, j, k) not in cls.wendland_coefficient_cache:
                if j == 0:
                    cls.wendland_coefficient_cache[(l, j, k)] = sum(
                        cls.wendland_coefficient(
                            l, j, k-1
                        ) / (j + 2)
                        for j in range(0, l + 2*k + 1)
                    )
                else:
                    cls.wendland_coefficient_cache[(l, j, k)] = -cls.wendland_coefficient(
                        l, j - 2, k - 1
                    ) / j
                # else:
                #     raise ValueError("bad ljk ({}, {}, {})".format(l, j, k))
            return cls.wendland_coefficient_cache[(l, j, k)]
    @classmethod
    def wendland_polynomial(cls, r, d=None, k=3, inds=None):
        l = int(np.floor(d/2) + k + 1)
        powers = np.arange(0, l+2*k+1)
        coeffs = [
            cls.wendland_coefficient(l, j, k)
                if j % 2 == 0 or j > 2 * k - 1 else # "the first k odd powers vanish"
            0
            for j in powers
        ]
        # raise Exception(coeffs)
        return np.dot(coeffs, r[np.newaxis, :] ** powers[:, np.newaxis]) * (r <= 1).astype(int)
    @classmethod
    def wendland_polynomial_derivative(cls, n: int, inds=None):
        def deriv(r, k=3, d=None, inds=inds):
            l = np.floor(d / 2) + k + 1
            powers = np.arange(n, l + 2*k+1)
            coeffs = [
                np.prod(j - np.arange(n)) * cls.wendland_coefficient(l, j, k)
                    if j % 2 == 1 or j > 2 * k - 1 else  # "the first k odd powers vanish"
                0
                for j in powers
            ]
            return np.dot(coeffs, r[np.newaxis, :] ** (powers-n)[:, np.newaxis]) * (r <= 1).astype(int)
        return deriv
    @staticmethod
    def wendland_polynomial_singularity_handler(n: int, ndim: int, inds=None):
        # handler for what happens at zero specifically
        def deriv(r, k=3, d=None, inds=inds):  # r will be a vector of zeros...
            if n - (1 - (k % 2)) > k:
                raise ValueError("Wendland polynomial deriv not defined")
            return np.zeros((len(r),) + (ndim,) * n)
        return deriv

    @staticmethod
    def zeros(r, inds=None):
        return np.zeros(len(r))
    @staticmethod
    def zeros_derivative(n: int, inds=None):
        def deriv(r, inds=inds):
            return np.zeros(len(r))
        return deriv
    @staticmethod
    def zeros_singularity_handler(n: int, ndim: int, inds=None):
        # handler for what happens at zero specifically
        def deriv(r, inds=inds):  # r will be a vector of zeros...
            return np.zeros((len(r),) + (ndim,) * n)
        return deriv

    @property
    def default_kernels(self):
        return {
            'gaussian': {
                'function': self.gaussian,
                'derivatives': self.gaussian_derivative,
                'zero_handler': self.gaussian_singularity_handler
            },

            'thin_plate_spline': {
                'function': self.thin_plate_spline,
                'derivatives': self.thin_plate_spline_derivative,
                'zero_handler': self.thin_plate_spline_singularity_handler
            },

            'wendland_polynomial': {
                'function': self.wendland_polynomial,
                'derivatives': self.wendland_polynomial_derivative,
                'zero_handler': self.wendland_polynomial_singularity_handler
            },

            'zeros': {
                'function': self.zeros,
                'derivatives': self.zeros_derivative,
                'zero_handler': self.zeros_singularity_handler
            }
        }

    @staticmethod
    def morse(r, a=1, inds=None):
        if inds is not None:
            if not isinstance(a, (int, float, np.integer, np.floating)):
                a = a[inds]
        return (1-np.exp(-a*r))
    @staticmethod
    def morse_derivative(n: int, inds=None):
        s = -1 if n % 2 == 1 else 1
        def deriv(r, a=1, inds=inds):
            if inds is not None:
                if not isinstance(a, (int, float, np.integer, np.floating)):
                    a = a[inds]
            return s * a**n * np.exp(-a*r) #TODO: I think this is wrong...?

        return deriv

    @staticmethod
    def even_powers(r, o=1, inds=None):
        return r**(2*o)
    @staticmethod
    def even_powers_deriv(n: int, inds=None):
        # s = -1 if n % 2 == 1 else 1

        def deriv(r, o=1, inds=inds):
            s = np.prod(np.arange(2*o, 2*o - n, -1))
            return s * r**(2*o - n)

        return deriv

    @staticmethod
    def laguerre(r, k=3, shift=2.29428, inds=None):
        return special.eval_laguerre(k, r + shift) # shift to the minimum basically...
    @staticmethod
    def laguerre_deriv(n:int, inds=None):
        """(-1)^n LaguerreL[k - n, n, x]"""
        s = -1 if n % 2 == 1 else 1
        def deriv(r, k=3, shift=2.29428, inds=inds):
            return s * special.eval_genlaguerre(k-n, n, r + shift)
        return deriv

    @classmethod
    def compact_laguerre(cls, r, e=1, k=3, shift=2.29428, inds=None):
        return cls.gaussian(r, e=e) * cls.laguerre(r, k=k, shift=shift)
    @classmethod
    def compact_laguerre_deriv(cls, n:int, inds=None):
        from scipy.special import hermite

        funs = [
            (
                cls.gaussian_derivative(i) if i > 0 else cls.gaussian,
                cls.laguerre_deriv(n-i) if i < n else cls.laguerre
            )
            for i in range(n+1)
        ]
        def deriv(r, e=1, k=3, shift=2.29428, inds=inds):
            val = 0
            for g,l in funs:
                val = val + g(r, e=e)*l(r, k=k, shift=shift)
            return val

        return deriv

    @property
    def default_auxiliary_bases(self):
        return {
            'power':None,
            'morse':{
                'function': self.morse,
                'derivatives': self.morse_derivative
            },
            'even_powers':{
                'function': self.even_powers,
                'derivatives': self.even_powers_deriv
            },
            'laguerre':{
                'function': self.laguerre,
                'derivatives': self.laguerre_deriv
            },
            'compact_laguerre':{
                'function': self.compact_laguerre,
                'derivatives': self.compact_laguerre_deriv
            }
        }

    def _poly_exprs(self, ndim:int, order:int, deriv_order:int, monomials=True):
        k = (ndim, order, deriv_order)
        if k not in self._expr_cache:
            if self.aux_poly['function'] is not None:
                coord_expr = TensorExpression.ScalarFunctionTerm(
                    TensorExpression.CoordinateVector(ndim, name='coord_vec'),
                    f=self.aux_poly
                )
            else:
                coord_expr = TensorExpression.CoordinateVector(ndim, name='coord_vec')
            if monomials:
                base_expr = TensorExpression.ScalarPowerTerm(coord_expr, order)
            else:
                # if self.aux_poly['function'] is not None:
                #     raise NotImplementedError(
                #         "auxiliary polynomial not supported for non-monomial basis".format(self.aux_poly)
                #     )
                base_expr = TensorExpression.OuterPowerTerm(coord_expr, order)
            exprs = [base_expr]
            for _ in range(deriv_order):
                exprs.append(exprs[-1].dQ())
            self._expr_cache[k] = exprs
        return self._expr_cache[k]
    def evaluate_poly_matrix(self, pts, degree, deriv_order=0, poly_origin=0.5, include_constant_term=True, monomials=True):

        if isinstance(poly_origin, (int, float, np.integer, np.floating)):
            pts = pts - poly_origin
        else:
            poly_origin = np.asanyarray(poly_origin)
            pts = pts[np.newaxis, :, :] - poly_origin[:, np.newaxis, :]

        ndim = pts.shape[-1]
        exprs = (
                (
                    [    # add in constant term by hand
                        [
                            TensorExpression.ConstantArray(np.ones(pts.shape[:-1] + (1,)))
                        ] +
                        [
                            TensorExpression.ConstantArray(np.zeros(pts.shape[:-1] + (ndim,)*d + (1,)) )
                            for d in range(1, deriv_order+1)
                        ]
                    ]
                    if include_constant_term else
                    []
                ) +
                [self._poly_exprs(ndim, d, deriv_order, monomials=monomials) for d in range(1, degree + 1)]
        )

        crds = TensorExpression.ArrayStack(pts.shape[:-1], pts)
        vals = TensorExpression(exprs, coord_vec=crds).eval()

        # raise Exception(
        #     pts.shape,
        #     [[v.shape for v in _] for _ in vals])

        # for exp_list in exprs:
        #     sublist = []
        #     for e in exp_list:
        #         print("=="*50)
        #         print(e)
        #         sublist.append(TensorExpression(e, coord_vec=crds).eval())
        #     vals.append(sublist)
        # vals = [ # evaluate poly values & derivs at each order in a dumb, kinda inefficient way? (for now)
        #     [
        #         TensorExpression(e, coord_vec=crds).eval()
        #         for e in exp_list
        #     ]
        #     for exp_list in exprs
        # ]

        # print("==="*20)
        # print(
        #     len(vals), len(vals[0]), vals[0][0].shape,
        #     len(old_vals), len(old_vals[0]), old_vals[0][0].shape
        # )
        # for new_v, old_v in zip(vals, old_vals):
        #     for a1, a2 in zip(new_v, old_v):
        #         print(a1.shape, a2.shape)
        #         print(np.sum(a1))
        #         print(np.sum(a2))
        #         assert np.allclose(a1, a2)
        #
        # raise ValueError("...")

        # Set things up so that arbitrary numbers of centers can be used at once
        padding = (slice(None, None, None),) * (2 if isinstance(poly_origin, np.ndarray) else 1)

        # now we take only upper triangle indices to get only unique ones
        flat_vals = [[] for _ in range(deriv_order+1)]
        for o, v_list in enumerate(vals): # loop over different stacks
            # get upper inds for the OuterPower term
            if not include_constant_term or o > 0:
                if not include_constant_term:
                    o = o+1
                if monomials:
                    power_inds = ()
                else:
                    power_inds = self.triu_inds(ndim, o)
                pis = padding + power_inds
                flat_vals[0].append(v_list[0][pis])
            else:
                power_inds = ()
                flat_vals[0].append(v_list[0]) # need to add a dimension for joining...
            for n,d in enumerate(v_list[1:]): #need utri inds
                # sample first power inds then derivs
                pis = padding + (slice(None, None, None),)*(n+1) + power_inds
                d = d[pis]
                # get upper inds for the different derivatives
                inds = self.triu_inds(ndim, n+1)
                inds = padding + inds
                d = d[inds]
                flat_vals[n+1].append(d)

        mats = [np.concatenate(v, axis=-1) for v in flat_vals]
        if isinstance(poly_origin, np.ndarray):
            mats = [np.moveaxis(m, 0, -1) for m in mats]
            mats = [np.reshape(m, m.shape[:-2] + (-1,)) for m in mats]
        mat = np.concatenate(
            [
                m.reshape((-1, m.shape[-1]))
                for m in mats
            ],
            axis=0
        )

        # max_value_cutoff = 1e20  # anything bigger is likely to cause numerical issues...
        # mat = np.nan_to_num(mat, nan=max_value_cutoff)
        # shit_pos = np.abs(mat) > max_value_cutoff
        # mat[shit_pos] = np.sign(mat[shit_pos]) * max_value_cutoff
        return mat

    def _rbf_exprs(self, ndim, deriv_order, inds):
        k = (ndim, deriv_order)
        # if k not in self._expr_cache:
        exprs = [
            TensorExpression.ScalarFunctionTerm(
                TensorExpression.VectorNormTerm(
                    TensorExpression.CoordinateVector(ndim, name="pts")
                ),
                f={
                    k:(lambda r,f=f:f(r, inds=inds))
                    for k,f in self.kernel.items()
                }
            )
        ]
        for _ in range(deriv_order):
            exprs.append(exprs[-1].dQ())
        self._expr_cache[k] = exprs
        return self._expr_cache[k]
    def evaluate_rbf_matrix(self, pts, centers, inds, deriv_order=0, zero_tol=1e-8):
        displacements = np.reshape(pts[:, np.newaxis, :] - centers[np.newaxis, :, :], (-1, centers.shape[-1]))
        inds = np.broadcast_to(inds[np.newaxis], (len(pts), len(centers))).flatten()
        distances = np.linalg.norm(displacements, axis=1)
        zero_pos = np.where(np.abs(distances) <= zero_tol) # we don't feed these displacements into the derivative evals
        nonz_pos = np.where(np.abs(distances) > zero_tol) # this could be fast but w/e
        displacements = displacements[nonz_pos]

        nonz_inds = inds[nonz_pos]
        zero_inds = inds[zero_pos]

        # now we evaluate the norm term...
        ndim = displacements.shape[-1]
        crds = TensorExpression.ArrayStack((len(displacements),), displacements)
        expr_vals = TensorExpression(self._rbf_exprs(ndim, deriv_order, nonz_inds), pts=crds).eval()
            # for e in self._rbf_exprs(ndim, deriv_order)
        # ]
        flat_expr_vals = [expr_vals[0]]
        for n, d in enumerate(expr_vals[1:]):  # need utri inds
            # get upper inds for the different derivatives
            inds = self.triu_inds(ndim, n+1)# tuple(np.array(list(itertools.combinations_with_replacement(range(ndim), r=n + 1))).T)
            inds = (slice(None, None, None),) + inds
            d = d[inds]
            flat_expr_vals.append(d)

        # handle singularities
        if len(zero_pos) > 0:
            zv = np.zeros(len(zero_pos[0]))
            # now we manage the singularities in the derivs
            sing_vals = [self.kernel['function'](zv, inds=zero_inds)]
            handler = self.kernel['zero_handler']
            for n in range(1, deriv_order+1):
               d = handler(n, ndim, inds=zero_inds)(zv)
               inds = self.triu_inds(ndim, n)#tuple(np.array(list(itertools.combinations_with_replacement(range(ndim), r=n))).T)
               inds = (slice(None, None, None),) + inds
               d = d[inds]
               sing_vals.append(d)

            #create final vectors by merge datasets appropriately
            full_exprs = []
            for v,s in zip(flat_expr_vals, sing_vals):
                a = np.zeros( (len(distances),) + v.shape[1:] )
                a[zero_pos] = s
                a[nonz_pos] = v
                full_exprs.append(a)
            flat_expr_vals = full_exprs

        # now reshape to turn into full mats
        npts = len(centers)

        flat_expr_vals = [
            np.moveaxis(
                a.reshape((len(pts), len(centers)) + a.shape[1:]),
                1, -1
            )
            for a in flat_expr_vals
        ]
        mat = np.concatenate(
            [
                a.reshape((-1, npts))
                for a in flat_expr_vals
            ],
            axis=0
        )

        return mat
    poly_origin = 0.5
    def construct_matrix(self,
                         pts,
                         centers,
                         inds,
                         degree=0, # passed independently
                         deriv_order=0, # passed independently
                         zero_tol=1e-8, # not really intended to be changed...
                         poly_origin=None, # can be set at the class level
                         include_constant_term=True, # can be set at the class level
                         force_square=False,
                         monomials=True,
                         multicentered_polys=False
                         ):
        if poly_origin is None:
            poly_origin = self.poly_origin
        if degree > 0:
            pol_mats = self.evaluate_poly_matrix(pts, degree,
                                                 poly_origin=poly_origin if not multicentered_polys else centers,
                                                 include_constant_term=include_constant_term,
                                                 deriv_order=deriv_order,
                                                 monomials=monomials)
        else:
            pol_mats = None
        rbf_mats = self.evaluate_rbf_matrix(pts, centers, inds, deriv_order=deriv_order, zero_tol=zero_tol)
        # print(">>", degree, rbf_mats.shape, pol_mats.shape)
        if pol_mats is not None:
            rbf_mats = np.concatenate([rbf_mats, pol_mats], axis=1)
        if force_square and rbf_mats.shape[1] > rbf_mats.shape[0]: # force it to be square by including extra polys
            curl = rbf_mats.shape[0]
            neel = rbf_mats.shape[1]
            disp = neel-curl
            rbf_mats = np.concatenate([rbf_mats, np.zeros((disp, neel))], axis=0)
            rbf_mats[curl:, :curl] = rbf_mats[:curl, -disp:].T
        return rbf_mats

    @staticmethod
    def svd_solve(a, b, svd_cutoff=1e-12):
        [U, s, Vt] = np.linalg.svd(a, full_matrices=False)
        r = max(np.where(s >= svd_cutoff)[0])
        temp = np.dot(U[:, :r].T, b) / s[:r]
        return np.dot(Vt[:r, :].T, temp)
    def solve_system(self, centers, vals, derivs:list, inds, solver=None, return_data=False, error_threshold=None):
        if len(vals) != len(centers):
            raise ValueError("need same number of interpolation points and values (got {} and {})".format(
                len(centers),
                len(vals)
            ))
        # TODO: compute the appropriate deriv order to match the vals and derivs
        derivs = [d.flatten() for d in derivs]
        nder = len(derivs)
        og_vals = vals
        vals = np.concatenate([vals]+derivs, axis=0)

        degree = 0
        if nder > 0:
            # I now know how many derivs I have per order...but I get the same contrib from each
            # degree order, but the issue is the num derivs scales with the number of centers too
            # so it needs to be solved for differently
            ndim = centers.shape[-1]
            target_extra = len(centers)*sum(self.triu_num(ndim, k) for k in range(1, nder + (0 if self.multicenter_monomials else 1) )) - 1
            cur = 0
            ncent = len(centers) if self.multicenter_monomials else 1
            for degree in range(1, nder*(target_extra//ndim)):# not sure where I need to stop?
                cur = cur + (self.triu_num(ndim, degree) if not self.monomial_basis else ndim*ncent)
                if cur >= target_extra:
                    break

        M = self.construct_matrix(centers, centers, inds,
                                  degree=degree+self.extra_degree,
                                  deriv_order=nder,
                                  monomials=self.monomial_basis,
                                  multicentered_polys=self.multicenter_monomials,
                                  force_square=True) # deriv_order and degree collaborate to get a square matrix...?

        if len(M) > len(vals):
            vals = np.concatenate([vals, np.zeros(len(M) - len(vals))])

        # with np.printoptions(linewidth=1e8, threshold=1e8):
        #     # print(centers)
        #     # print(np.min(M), np.max(M), np.linalg.cond(M))
        #     # print(vals[:, np.newaxis])
        if self.max_condition_number is not None and self.max_condition_number is not np.inf:
            cn = np.linalg.cond(M)
            if cn > self.max_condition_number:
                print("WARNING: condition number={}".format(cn))

        if solver is None:
            solver = self.solve_method
        if solver == "least_squares":
            # self.logger.log_print("   dimension of interpolation matrix: {}".format(M.shape))
            w_data = np.linalg.lstsq(M, vals, rcond=None)
        elif solver == "svd":
            w_data = self.svd_solve(M, vals)  # , rcond=None)
            w_data = (w_data,)
        elif solver is None or solver == 'solver':
            w_data = np.linalg.solve(M, vals)#, rcond=None)
            w_data = (w_data,)

        # with np.printoptions(linewidth=1e8, threshold=1e8):
        #     print(w_data[0])
        #     print(
        #         len(w_data[0]),
        #         np.argmin(w_data[0]), np.max(w_data[0]),
        #         np.argmax(w_data[0]), np.min(w_data[0]),
        #         np.std(w_data)
        #     )

        test_vals = np.dot(M, w_data[0])
        nv = len(og_vals)
        base_error = og_vals - test_vals[:nv]
        # often there's a consistent shift when things break down
        extra_shift = np.average(base_error)

        evec = np.concatenate([
            base_error - extra_shift,
            (vals[nv:] - test_vals[nv:])
        ])
        test_norm = np.linalg.norm(evec) / len(evec)  # average percent error ?
        # print(">>", test_norm, error_threshold)
        # print(len(evec), len(vals))

        test_err = np.max(np.abs(evec)) # max diff error?

        errors = [test_norm, test_err]

        if error_threshold is None:
            error_threshold = self.error_threshold

        if error_threshold is not None:
            if test_norm > error_threshold: #
                # print("...?")
                raise RBFDError("solution doesn't solve equations")
                # print("="*5)
                # print(test_norm)
                # print(vals - test_vals)



        if return_data:
            data = {'matrix':M, 'vals':vals, 'condition_number':cn}
            return w_data, degree+self.extra_degree, extra_shift, errors, data
        else:
            return w_data, degree+self.extra_degree, extra_shift, errors

    class InterpolationData:
        __slots__ = ['weights', 'centers', 'degree', 'scaling_data', 'extra_shift', 'interpolation_error', 'solver_data']
        def __init__(self, w, grid, degree, scaling_data, extra_shift=0, interpolation_error=0, solver_data=None):
            self.weights = w
            self.centers = grid
            self.degree = degree
            self.scaling_data = scaling_data
            self.solver_data = solver_data
            self.extra_shift = extra_shift
            self.interpolation_error = interpolation_error
        # def eval(self, *args, **kwargs):
        #     grid = self.grid[vals_inds]
        #     vals = self.vals[vals_inds]
        #     derivs = [d[vals_inds] for d in self.derivs]
        #
        #     grid, vals, derivs, scaling_data = self.RescalingData.initialize_subgrid_data(
        #         grid,
        #         vals,
        #         derivs
        #     )
        #
        #     # print(vals_inds)
        #     # print(vals)
        #     # print(derivs)
        #
        #     w, degree = self.solve_system(grid, vals, derivs)
        #
        #     self._pos_cache[k] = (w, grid, degree, scaling_data)

    def construct_evaluation_matrix(self, pts, data, deriv_order=0):
        """
        :param pts:
        :type pts:
        :param data:
        :type data:
        :param deriv_order:
        :type deriv_order:
        :return:
        :rtype:
        """

        scaling_data = data.scaling_data
        # w = data.weights
        degree = data.degree
        centers = data.centers

        sub_pts = scaling_data.apply_renormalization(pts)
        M = self.construct_matrix(sub_pts, centers, degree=degree, deriv_order=deriv_order, monomials=self.monomial_basis, multicentered_polys=self.multicenter_monomials)
        return M

    def apply_interpolation(self, pts, data, inds, reshape_derivatives=True, return_data=False, deriv_order=0):
        """

        :param pts:
        :type pts:
        :param data:
        :type data:
        :param deriv_order:
        :type deriv_order:
        :return:
        :rtype:
        """

        scaling_data = data.scaling_data
        w = data.weights
        degree = data.degree
        centers = data.centers

        sub_pts = scaling_data.apply_renormalization(pts)
        M = self.construct_matrix(sub_pts, centers, inds, degree=degree,
                                  deriv_order=deriv_order,
                                  monomials=self.monomial_basis,
                                  multicentered_polys=self.multicenter_monomials
                                  )
        # with np.printoptions(linewidth=1e8, threshold=1e8):
        #     print(sub_pts, M[:, -10:])
        v = np.dot(M, w[0])

        npts = len(sub_pts)
        vals = scaling_data.reverse_value_renormalization(v[:npts] - data.extra_shift)
        dvs = []
        if deriv_order > 0:  # derivatives
            offset = npts
            ndim = pts.shape[-1]
            for o in range(1, 1 + deriv_order):
                num = npts * self.triu_num(ndim, o)
                dvs.append(v[offset:offset + num])
                # print(dvs[-1])
                offset += num
            dvs = scaling_data.reverse_derivative_renormalization(dvs, reshape=reshape_derivatives)

        res = (vals, dvs)
        if return_data:
            res = res + (M,)
        return res

    def construct_interpolation(self, inds, solver_data=False, return_error=False):
        grid, vals, derivs, scaling_data = self.prep_interpolation_data(inds)

        # print("< ", grid.shape, vals.shape, [d.shape for d in derivs])

        res = self.solve_system(grid, vals, derivs, inds, return_data=solver_data)
        if solver_data:
            w, degree, extra, error, data = res
        else:
            w, degree, extra, error = res
            data = None

        #TODO: should I be tracking the indices of the points as well?
        # currently we just let those disappear but it's entirely possible
        # that someone might want to do some extra testing (a la resiliance test)
        # which would make it relevant to have access to extra information
        # relating to the original data...?
        new = self.InterpolationData(w, grid, degree, scaling_data, extra_shift=extra, interpolation_error=error, solver_data=data)
        return new

    class Interpolator(NeighborBasedInterpolator.Interpolator):
        parent:'RBFDInterpolator'
        def matrix(self, pts, deriv_order=0):
            pts = np.asanyarray(pts)
            smol = pts.ndim == 1
            if smol:
                pts = pts[np.newaxis]
            return self.parent.construct_evaluation_matrix(pts, self.data, deriv_order=deriv_order)

class DistanceWeightedInterpolator(NeighborBasedInterpolator):
    """
    Provides a quick implementation of inverse distance weighted interpolation
    """

    class InterpolationData:
        __slots__ = ['centers']
        def __init__(self, grid):
            self.centers = grid

    def construct_interpolation(self, inds, solver_data=False, return_error=False):
        grid = self.grid[inds]
        new = self.InterpolationData(grid)
        return new

    def apply_weights(self, weights, inds, deriv_order=0, reshape_derivatives=False):
        """
        Weights the values and derivatives

        :param weights: npts x hood_size matrix of weights
        :type weights:
        :param inds: npts x hood_size set of indices for the values
        :type inds:
        :return:
        :rtype:
        """
        vals = self.vals[inds]
        derivs = [self.derivs[i][inds] for i in range(deriv_order)]

        vals = np.matmul(weights[:, np.newaxis], vals[:, :, np.newaxis]).reshape((-1,))
        derivs = [
            vec_tensordot(d, weights, shared=1, axes=[1, 1]).reshape((-1,) + d.shape[2:])
            for d in derivs
        ]

        if reshape_derivatives:
            npts = len(vals)
            ndim = self.grid.shape[-1]
            reshaped = []

            for k,values in enumerate(derivs):
                k = k + 1
                shp = (npts,) + (ndim,) * k
                if shp != values.shape:
                    a = np.zeros(shp)
                    base_inds = NeighborBasedInterpolator.triu_inds(ndim, k)
                    inds = (slice(None, None, None),) + base_inds
                    a[inds] = values.reshape((npts, -1))
                    for pos in np.array(base_inds).T:
                        base = (slice(None, None, None),) + tuple(pos)
                        for p in itertools.permutations(pos):
                            ind = (slice(None, None, None),) + p
                            a[ind] = a[base]
                    values = a
                reshaped.append(values)

        return vals, derivs

    @abc.abstractmethod
    def get_weights(self, pts, data, inds):
        """
        Computes weights for pts from the inds

        :param pts:
        :type pts:
        :param inds:
        :type inds:
        :return:
        :rtype:
        """
        ...

    def apply_interpolation(self,
                            pts, data, inds,
                            deriv_order=0,
                            reshape_derivatives=True,
                            return_data=False
                            ):
        """
        :param pts:
        :type pts:
        :param data:
        :type data:
        :param deriv_order:
        :type deriv_order:
        :return:
        :rtype:
        """

        weights = self.get_weights(pts, data, inds)

        vals, derivs = self.apply_weights(weights, inds, deriv_order=deriv_order, reshape_derivatives=reshape_derivatives)

        if return_data:
            return vals, derivs, weights
        else:
            return vals, derivs

class InverseDistanceWeightedInterpolator(DistanceWeightedInterpolator):

    @classmethod
    def weight_deriv(cls, disp, dists, norm, power, n, gammas_1=None):
        prefactor = power + n + 1
        if gammas_1 is None: # we can reuse this...
            gammas_1 = dists ** -(power + 2*n) / norm[:, np.newaxis]
        gammas_2 = dists ** -(power + 2*(n+1)) / norm[:, np.newaxis]
        gouter = gammas_1[:, :, np.newaxis] * gammas_2[:, np.newaxis, :] # N x k x k
        grad = prefactor * (disp - np.matmul(gouter, disp)).transpose(0, 2, 1) # disp is N x k x d
        return gammas_2, grad

    @classmethod
    def idw_derivs(cls, deriv_order, disp, dists, norm, power, weights):
        derivs = []
        gammas = []
        if deriv_order > 0:
            gammas1, grad = cls.weight_deriv(disp, dists, norm, power, 0, weights)
            gammas.append(gammas1)
            derivs.append(grad)
        if deriv_order > 1:
            gammas1 = gammas[0]
            grad = derivs[0]
            gammas2, grad2 = cls.weight_deriv(disp, dists, norm, power, 1, gammas1)
            gammas.append(gammas2)
            prefactor = power + 2
            g1_expansion = [gammas1, grad]
            g2_expansion = [gammas2, grad2]
            inner_term_expansion = [
                vec_tensordiag(gammas1, axis=1) - vec_outer(gammas1, gammas2, axes=[-1, -1]),
                vec_tensordiag(grad, axis=2) - tensorprod_deriv(g1_expansion, g2_expansion, [1], axes=[-1, -1])[0]
            ]
            base_eye = np.eye(disp.shape[-1])
            disp_eye = np.broadcast_to(base_eye[np.newaxis], disp.shape[:2] + base_eye.shape)
            disp_eye = np.moveaxis(disp_eye, 2, 1)
            disp_expansion = [disp, disp_eye]
            # print("???", [e.shape for e in inner_term_expansion], [d.shape for d in disp_expansion])
            hess = prefactor * tensordot_deriv(inner_term_expansion, disp_expansion, [1],
                                               axes=[2, 1], shared=1
                                               )[0]
            hess = np.moveaxis(hess, 2, -1)
            derivs.append(hess)

        if deriv_order > 2:
            raise NotImplementedError('idw_derivs only up to 2nd order')
        return derivs

    @classmethod
    def get_idw_weights(cls, pts, dists, disps=None, deriv_order=None, zero_tol=1e-6, power=2):
        scaling = np.min(dists, axis=1)
        zero_pos = np.where(dists < zero_tol)
        if len(zero_pos) > 0 and len(zero_pos[0]) > 0:
            weights = np.zeros_like(dists)
            zero_row, zero_idx = np.unique(zero_pos[0], return_index=True) # so annoying but have to be clean about degenerate abcissae
            zero_col = zero_pos[1][zero_idx]
            weights[zero_row, zero_col] = 1
            nonz_pos = np.setdiff1d(np.arange(len(pts)), zero_row)
            if len(nonz_pos) > 0:
                normed_dists = dists[nonz_pos] / scaling[nonz_pos, np.newaxis]
                sdists = np.power(normed_dists, -power)
                norm = np.sum(sdists, axis=1)
                weights[nonz_pos] = sdists / norm[:, np.newaxis]
                if deriv_order is not None:
                    if disps is None: raise ValueError("displacements needed for derivatives")
                    weight_derivs = cls.idw_derivs(deriv_order, disps[nonz_pos], normed_dists, norm,
                                                   power, weights[nonz_pos])
                    derivs = []
                    for w in weight_derivs:
                        full_d = np.zeros((len(dists),) + w.shape[1:])
                        full_d[nonz_pos] = w
                        derivs.append(full_d)
                    weights = [weights] + derivs
            elif deriv_order is not None:
                weights = [weights]
        else:
            normed_dists = dists / scaling[:, np.newaxis]
            sdists = np.power(normed_dists, -power)
            norm = np.sum(sdists, axis=1)
            weights = sdists / norm[:, np.newaxis]
            if deriv_order is not None:
                if disps is None: raise ValueError("displacements needed for derivatives")
                weight_derivs = cls.idw_derivs(deriv_order, disps, normed_dists, norm, power, weights)
                weights = [weights] + weight_derivs

        return weights
    def get_weights(self, pts, dists, inds, zero_tol=1e-6, power=2):
        return self.get_idw_weights(pts, dists, zero_tol=zero_tol, power=power)

    def eval(self,
             pts,
             deriv_order=0,
             neighbors=None,
             merge_neighbors=None,
             reshape_derivatives=True,
             return_interpolation_data=False,
             check_in_sample=True,
             zero_tol=1e-8,
             return_error=False,
             use_cache=True,
             retries=None,
             # resiliance_test_options=None,
             # extrapolation_warning=None
             max_distance=None,
             min_distance=None,
             neighborhood_clustering_radius=None,
             use_natural_neighbors=False,
             chunk_size=None,
             power=2,
             mode='fast'
             ):
        if deriv_order is None:
            deriv_order = 0

        if mode == 'fast':
            pts = np.asanyarray(pts)
            # check against coord shape ?
            # if pts.shape[-len(self._coord_shape):] == self._coord_shape:
            #     pts.reshape()
            if pts.ndim == 1:
                pts = pts[np.newaxis]

            if pts.ndim > 2:
                extra_shape = pts.shape[:-1]
                pts = pts.reshape(-1, pts.shape[-1])
            else:
                extra_shape = None

            hoods, dists = self.get_neighborhood(pts, neighbors=neighbors, return_distances=True)
            val_sets, der_sets = self.apply_interpolation(pts, dists, hoods,
                                                          deriv_order=deriv_order,
                                                          reshape_derivatives=reshape_derivatives)

            if extra_shape is not None:
                val_sets = np.reshape(val_sets, extra_shape)
                # if return_interpolation_data:
                #     interp_data = np.reshape(weights, extra_shape) # wat
                # if return_error:
                #     errors = np.reshape(errors, extra_shape + (2,)) # wat
                if der_sets is not None:
                    der_sets = [
                        d.reshape(extra_shape + d.shape[1:])
                        for d in der_sets
                    ]

            if deriv_order > 0:
                res = [val_sets] + der_sets
            else:
                res = val_sets

            # if return_interpolation_data or return_error:
            #     res = (res,)
            #
            # if return_error:
            #     res = res + (errors,)

            # if return_interpolation_data:
            #     res = res + (interp_data,)

            return res
        else:
            return super().eval(
                pts,
                deriv_order=deriv_order,
                neighbors=neighbors,
                merge_neighbors=merge_neighbors,
                reshape_derivatives=reshape_derivatives,
                return_interpolation_data=return_interpolation_data,
                check_in_sample=check_in_sample,
                zero_tol=zero_tol,
                return_error=return_error,
                use_cache=use_cache,
                retries=retries,
                # resiliance_test_options=None,
                # extrapolation_warning=None
                max_distance=max_distance,
                min_distance=min_distance,
                neighborhood_clustering_radius=neighborhood_clustering_radius,
                use_natural_neighbors=use_natural_neighbors,
                chunk_size=chunk_size
            )

# class NaturalNeighborInterpolator(NeighborBasedInterpolator):
#     """
#     Provides a quick implementation of natural neighbor interpolation
#     """
#     ...


import os, shutil
from string import Formatter
from .FileMatcher import *

__all__ = [
    "TemplateWriter",
    "OptionalTemplate"
]

class TemplateWriter:
    """
    A general class that can take a directory layout and apply template parameters to it
    Very unsophisticated but workable. For a more sophisticated take that walks through
    object trees, see `TemplateEngine`.
    """

    ignored_files = [".DS_Store"]

    def __init__(self, template_dir, replacements = None, file_filter = None, **opts):
        if replacements is not None:
            opts = replacements
        self._rep_dict = opts
        self._reps = None

        if file_filter is None:
            file_filter = FileMatcher(MatchList(*self.ignored_files, negative_match=True), use_basename=True)
        self.filter = FileMatcher(file_filter) if not isinstance(file_filter, StringMatcher) else file_filter
        self.template_dir = os.path.abspath(template_dir)

    @property
    def replacements(self):
        if self._reps is None:
            self._reps =  tuple(("`"+k+"`", str(v)) for k,v in self._rep_dict.items())
        return self._reps

    def apply_replacements(self, string):
        """Applies the defined replacements to the

        :param string:
        :type string:
        :return:
        :rtype:
        """
        import functools

        return functools.reduce(lambda s, r: s.replace(*r), self.replacements, string)

    def write_file(self, template_file, out_dir, apply_template = True, template_dir = None):
        """writes a single _file_ to _dir_ and fills the template from the parameters passed when intializing the class

        :param template_file: the file to load and write into
        :type template_file: str
        :param out_dir: the directory to write the file into
        :type out_dir: str
        :param apply_template: whether to apply the template parameters to the file content or not
        :type apply_template: bool
        :return:
        :rtype:
        """

        if template_dir is None:
            template_dir = self.template_dir

        stripped_loc = template_file.split(template_dir, 1)[-1]
        rep_loc = self.apply_replacements(stripped_loc).strip(os.sep)
        new_file = os.path.join(out_dir, rep_loc)

        os.makedirs(os.path.dirname(new_file), exist_ok = True)

        if apply_template:
            with open(template_file) as tf:
                try:
                    content = tf.read()
                except:
                    raise IOError("Couldn't read content from {}".format(template_file))
                with open(new_file, "w+") as out:
                    out.write(self.apply_replacements(content))
        else:
            shutil.copy(template_file, new_file)

        return new_file

    def iterate_write(self, out_dir, apply_template = True, src_dir = None, template_dir = None):
        """Iterates through the files in the template_dir and writes them out to dir

        :return:
        :rtype:
        """

        if template_dir is None:
            template_dir = self.template_dir
        if src_dir is None:
            src_dir = template_dir

        files = os.listdir(src_dir)
        for f in files:
            f = os.path.join(src_dir, f)
            if self.filter.matches(f.split(template_dir, 1)[-1]):
                if not os.path.isdir(f):
                    self.write_file(f, out_dir, apply_template = apply_template, template_dir = template_dir)
                else:
                    self.iterate_write(out_dir, apply_template = apply_template, src_dir = f, template_dir = template_dir)


class OptionalTemplate:
    def __init__(self, template, **opts):
        self.opts = opts
        if os.path.isfile(template):
            with open(template) as t:
                template = t.read()
        self.template = template

    class DefaultFormatter(Formatter):
        default_key = "-MISSING-"

        def get_value(self, key, args, kwds):
            if isinstance(key, str):
                try:
                    return kwds[key]
                except KeyError:
                    return self.default_key
            else:
                return Formatter.get_value(key, args, kwds)

    @classmethod
    def apply_template(cls, template, opts, formatter=None, strip_missing_blocks=None, strip_missing=True):
        if formatter is None:
            formatter = cls.DefaultFormatter()
        applied_template = formatter.format(template, **opts)
        if strip_missing_blocks is None: strip_missing_blocks = strip_missing
        if strip_missing_blocks:
            applied_template = applied_template.replace(formatter.default_key + "\n", "")
        if strip_missing:
            applied_template = applied_template.replace(formatter.default_key, "")
        return applied_template

    def apply(self, **opts):
        return self.apply_template(self.template, dict(self.opts, **opts))


"""
A job management package to make it easier to instantiate
job
"""

import time, datetime, os, shutil
from .Persistence import PersistenceManager
from .Checkpointing import JSONCheckpointer
from .Logging import Logger, NullLogger

from ..Parallelizers import Parallelizer

__all__ = [
    "Job",
    "JobManager"
]

class Job:
    """
    A job object to support simplified run scripting.
    Provides a `job_data` checkpoint file that stores basic
    data about job runtime and stuff, as well as a `logger` that
    makes it easy to plug into a run time that supports logging
    """
    default_job_file = "job_data.json"
    default_log_file = "log.txt"
    def __init__(self,
                 job_dir,
                 job_file=None,
                 logger=None,
                 parallelizer=None,
                 job_parameters=None
                 ):
        self.dir = job_dir
        self.end = None

        self.checkpoint = self.load_checkpoint(job_file)
        self.logger = self.load_logger(logger)
        self.parallelizer = self.load_parallelizer(parallelizer)

        self._init_dir = None

        self.job_parameters = job_parameters

    @classmethod
    def from_config(cls,
                    config_location=None,
                    job_file=None,
                    logger=None,
                    parallelizer=None,
                    job_parameters=None
                    ):
        return cls(config_location,
                   job_file=job_file,
                   logger=logger,
                   parallelizer=parallelizer,
                   job_parameters=job_parameters
                   )

    def load_checkpoint(self, job_file):
        """
        Loads the checkpoint we'll use to dump params

        :param job_file:
        :type job_file:
        :return:
        :rtype:
        """
        if job_file is None:
            job_file = self.default_job_file
        if os.path.abspath(job_file) != job_file:
            job_file = os.path.join(self.dir, job_file)
        return JSONCheckpointer(job_file)
    def load_logger(self, log_spec):
        """
        Loads the appropriate logger

        :param log_spec:
        :type log_spec: str | dict
        :return:
        :rtype:
        """
        if log_spec is None:
            log_spec = self.default_log_file
        if log_spec is True:
            return Logger()
        elif log_spec is False:
            return NullLogger()
        elif isinstance(log_spec, str):
            if os.path.abspath(log_spec) != log_spec:
                log_spec = os.path.join(self.dir, log_spec)
            return Logger(log_spec)
        elif isinstance(log_spec, dict):
            if 'log_file' not in log_spec:
                log_spec['log_file'] = self.default_log_file
            if os.path.abspath(log_spec['log_file']) != log_spec['log_file']:
                log_spec['log_file'] = os.path.join(self.dir, log_spec['log_file'])
            return Logger(**log_spec)
        else:
            raise ValueError("don't know what to do with log spec {}".format(
                log_spec
            ))

    def load_parallelizer(self, par_spec):
        """
        Loads the appropriate parallelizer.
        If something other than a dict is passed,
        tries out multiple specs sequentially until it finds one that works

        :param log_spec:
        :type log_spec: dict
        :return:
        :rtype:
        """
        if par_spec is None:
            return None
        elif isinstance(par_spec, dict):
            try:
                par = Parallelizer.from_config(**par_spec)
            except (ImportError, ModuleNotFoundError):
                return None
            else:
                return par
        else:
            for spec in par_spec:
                par = self.load_parallelizer(spec)
                if par is not None:
                    return par

    def path(self, *parts):
        """
        :param parts:
        :type parts: str
        :return:
        :rtype:
        """
        return os.path.join(self.dir, *parts)

    @property
    def working_directory(self):
        if 'working_directory' in self.job_parameters:
            init_dir = os.getcwd()
            try:
                os.chdir(self.dir)
                return os.path.abspath(self.job_parameters['working_directory'])
            finally:
                os.chdir(init_dir)
        else:
            return self.dir

    def __enter__(self):
        self._init_dir = os.getcwd()
        os.chdir(self.dir)
        self.checkpoint.__enter__()
        self.start = time.time()
        self.checkpoint['start'] = {
            'datetime': datetime.datetime.now().isoformat(),
            'timestamp': self.start
        }
        self.checkpoint['parameters'] = self.job_parameters
        if self.parallelizer is None:
            self.parallelizer = Parallelizer.get_default()
        self.parallelizer.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._init_dir)
        end = time.time()
        self.checkpoint['runtime'] = end - self.start
        self.checkpoint.__exit__(exc_type, exc_val, exc_tb)
        self.parallelizer.__exit__(exc_type, exc_val, exc_tb)

class JobManager(PersistenceManager):
    """
    A class to manage job instances.
    Thin layer on a `PersistenceManager`
    """
    default_job_type=Job
    def __init__(self, job_dir, job_type=None):
        if job_type is None:
            job_type = self.default_job_type
        super().__init__(job_type, persistence_loc=job_dir)

    def job(self, name, timestamp=False, **kw):
        """
        Returns a loaded or new job with the given name and settings

        :param name:
        :type name: str
        :param timestamp:
        :type timestamp:
        :param kw:
        :type kw:
        :return:
        :rtype: Job
        """
        if os.path.isdir(name):
            # kw['initialization_directory'] = name
            name = os.path.basename(name)
        if timestamp:
            name += "_" + datetime.datetime.now().isoformat()
        return self.load(name, make_new=True, init=kw)

    @classmethod
    def job_from_folder(cls, folder, job_type=None, make_config=True, **opts):
        """
        A special case convenience function that goes
        directly to starting a job from a folder

        :return:
        :rtype: Job
        """

        if make_config:
            test_file = os.path.join(folder, "config.json")
            if not os.path.exists(test_file):
                import json
                with open(test_file, "w+") as dump:
                    json.dump({}, dump)
        jm = cls(os.path.dirname(folder), job_type=job_type)
        return jm.job(folder, **opts)

    @classmethod
    def current_job(cls, job_type=None, make_config=True, **opts):
        """
        A special case convenience function that starts a
        JobManager one directory up from the current
        working directory and intializes a job from the
        current working directory

        :return:
        :rtype: Job
        """

        return cls.job_from_folder(os.getcwd(), job_type=job_type, make_config=make_config, **opts)






"""
Provides utilities to work with pretty low-level stuff.
Any lowish-level numerical operations that need to be shared go here.
"""

__all__ = []
from .EulerSystem import *; from .EulerSystem import __all__ as _all
__all__ += _all
from .TransformationMatrices import *; from .TransformationMatrices import __all__ as _all
__all__ += _all
from .TransformationTransformations import *; from .TransformationTransformations import __all__ as _all
__all__ += _all
from .VectorOps import *; from .VectorOps import __all__ as _all
__all__ += _all
from .CoordOps import *; from .CoordOps import __all__ as _all
__all__ += _all
from .Sparse import *; from .Sparse import __all__ as _all
__all__ += _all
from .SetOps import *; from .SetOps import __all__ as _all
__all__ += _all
from .Misc import *; from .Misc import __all__ as _all
__all__ += _all
from .TensorDerivatives import *; from .TensorDerivatives import __all__ as _all
__all__ += _all
from .CoordinateFrames import *; from .CoordinateFrames import __all__ as _all
__all__ += _all
from .Optimization import *; from .Optimization import __all__ as _all
__all__ += _all
from .GeometricTransformations import *; from .GeometricTransformations import __all__ as _all
__all__ += _all
from .PermutationOps import *; from .PermutationOps import __all__ as _all
__all__ += _all